from __future__ import annotations

from base64 import b32decode, b64decode
from datetime import UTC, datetime, timedelta
from typing import Annotated, Any

import pytest
from cryptography.hazmat.primitives.asymmetric.ed25519 import (
    Ed25519PrivateKey,
    Ed25519PublicKey,
)
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey
from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PrivateKey, X25519PublicKey
from pydantic import BaseModel, BeforeValidator, TypeAdapter, ValidationError

from aiostem.structures import AuthMethod
from aiostem.utils import (
    Base32Encoder,
    Base64Encoder,
    EncodedBytes,
    TrAfterAsTimezone,
    TrBeforeSetToNone,
    TrBeforeStringSplit,
    TrBeforeTimedelta,
    TrBoolYesNo,
    TrCast,
    TrEd25519PrivateKey,
    TrEd25519PublicKey,
    TrRSAPrivateKey,
    TrRSAPublicKey,
    TrX25519PrivateKey,
    TrX25519PublicKey,
)


class TestAsTimezone:
    @pytest.mark.parametrize(
        ('raw', 'timezone', 'timestamp'),
        [
            ('2024-12-09T23:10:14+01:00', None, 1733782214),
            ('2024-12-09T23:10:14', UTC, 1733785814),
        ],
    )
    def test_astimezone(self, raw, timezone, timestamp):
        adapter = TypeAdapter(Annotated[datetime, TrAfterAsTimezone(timezone)])
        result = adapter.validate_python(raw)
        assert int(result.timestamp()) == timestamp

    @pytest.mark.parametrize(
        'type_',
        [
            Annotated[int, TrAfterAsTimezone()],
            Annotated[None, TrAfterAsTimezone()],
        ],
    )
    def test_usage_error_on_source_type(self, type_):
        with pytest.raises(TypeError, match='source type is not a datetime'):
            TypeAdapter(type_)

    def test_json_schema(self):
        adapter = TypeAdapter(Annotated[datetime, TrAfterAsTimezone()])
        assert adapter.json_schema() == {'format': 'date-time', 'type': 'string'}


class TestSetToNone:
    ADAPTER_COMPLEX = TypeAdapter(
        Annotated[
            str | None,
            TrBeforeSetToNone({'NULL'}),
            TrBeforeSetToNone({'NIL'}),
        ]
    )
    ADAPTER_SIMPLE = TypeAdapter(
        Annotated[
            str | None,
            TrBeforeSetToNone({'NULL', 'NIL'}),
        ]
    )

    @pytest.mark.parametrize(
        ('entry', 'result'),
        [
            (None, None),
            ('NULL', None),
            ('NIL', None),
            ('ERROR', 'ERROR'),
        ],
    )
    def test_complex(self, entry, result):
        parsed = self.ADAPTER_COMPLEX.validate_python(entry)
        assert parsed == result

    @pytest.mark.parametrize(
        ('entry', 'result'),
        [
            (None, None),
            ('NULL', None),
            ('NIL', None),
            ('ERROR', 'ERROR'),
        ],
    )
    def test_simple(self, entry, result):
        parsed = self.ADAPTER_SIMPLE.validate_python(entry)
        assert parsed == result


class HostPort(BaseModel):
    host: str
    port: int


class TestStringSplit:
    @pytest.mark.parametrize(
        'entry',
        [
            [1, 2, 3, 4],
            '1,2,3,4',
            b'1,2,3,4',
        ],
    )
    def test_with_simple_types_always(self, entry: Any):
        adapter = TypeAdapter(Annotated[list[int], TrBeforeStringSplit()])
        res = adapter.validate_python(entry)
        assert res == [1, 2, 3, 4]

        assert adapter.dump_python(res) == '1,2,3,4'
        assert adapter.dump_json(res) == b'"1,2,3,4"'

    @pytest.mark.parametrize(
        'entry',
        [
            [1, 2, 3, 4],
            '1,2,3,4',
            b'1,2,3,4',
        ],
    )
    def test_with_simple_types_json(self, entry: Any):
        adapter = TypeAdapter(Annotated[list[int], TrBeforeStringSplit(when_used='json')])
        res = adapter.validate_python(entry)
        assert res == [1, 2, 3, 4]

        assert adapter.dump_python(res) == [1, 2, 3, 4]
        assert adapter.dump_json(res) == b'"1,2,3,4"'

    @pytest.mark.parametrize(
        ('entry', 'output'),
        [
            ('COOKIE', [AuthMethod.COOKIE]),
            ('NULL,SAFECOOKIE', [AuthMethod.NULL, AuthMethod.SAFECOOKIE]),
        ],
    )
    def test_with_strenum(self, entry: str, output: list[AuthMethod]):
        adapter = TypeAdapter(Annotated[list[AuthMethod], TrBeforeStringSplit()])
        for item in (entry, output):
            result = adapter.validate_python(item)
            assert result == output
            serial = adapter.dump_python(result)
            assert serial == entry

    def test_with_max_split(self):
        value = 'A,B,C,D'
        adapter = TypeAdapter(Annotated[list[str], TrBeforeStringSplit(maxsplit=1)])
        result = adapter.validate_python(value)
        assert len(result) == 2
        assert result[1] == 'B,C,D'

    @pytest.mark.parametrize(
        ('entry', 'serial'),
        [
            ('localhost:443', 'localhost:443'),
            (
                HostPort(host='localhost', port=443),
                'localhost:443',
            ),
        ],
    )
    def test_with_dict_keys(self, entry, serial):
        adapter = TypeAdapter(
            Annotated[
                HostPort,
                TrBeforeStringSplit(
                    dict_keys=('host', 'port'),
                    maxsplit=1,
                    separator=':',
                ),
            ]
        )
        result = adapter.validate_python(entry)
        assert isinstance(result, HostPort)
        assert result.host == 'localhost'
        assert result.port == 443

        serialized = adapter.dump_python(result)
        assert serialized == serial

    @pytest.mark.parametrize(
        'type_',
        [
            Annotated[int, TrBeforeStringSplit()],
            Annotated[None, TrBeforeStringSplit()],
        ],
    )
    def test_usage_error_as_sequence(self, type_):
        with pytest.raises(TypeError, match='source type is not a collection'):
            TypeAdapter(type_)


class TestTimedelta:
    ADAPTER_MINS = TypeAdapter(
        Annotated[
            timedelta,
            TrBeforeTimedelta(unit='minutes', is_float=False),
        ],
    )
    ADAPTER_SECS = TypeAdapter(Annotated[timedelta, TrBeforeTimedelta(unit='seconds')])
    ADAPTER_MSECS = TypeAdapter(Annotated[timedelta, TrBeforeTimedelta(unit='milliseconds')])

    @pytest.mark.parametrize(
        'entry',
        [
            timedelta(minutes=90),
            '01:30:00',
            '90',
            90,
        ],
    )
    def test_minutes_with_multiple_types(self, entry: Any):
        delta = self.ADAPTER_MINS.validate_python(entry)
        assert int(delta.total_seconds()) == 5400

        serial = self.ADAPTER_MINS.dump_python(delta)
        assert isinstance(serial, int)
        assert int(serial) == 90

    @pytest.mark.parametrize(
        'entry',
        [
            timedelta(seconds=1234),
            '00:20:34',
            '1234',
            1234,
        ],
    )
    def test_seconds_with_multiple_types(self, entry: Any):
        delta = self.ADAPTER_SECS.validate_python(entry)
        assert int(delta.total_seconds()) == 1234

        serial = self.ADAPTER_SECS.dump_python(delta)
        assert isinstance(serial, float)
        assert int(serial) == 1234

    @pytest.mark.parametrize(
        'entry',
        [
            timedelta(seconds=1.234),
            '00:00:01.234',
            '1234',
            1234,
        ],
    )
    def test_milliseconds_with_multiple_types(self, entry: Any):
        delta = self.ADAPTER_MSECS.validate_python(entry)
        assert int(delta.total_seconds()) == 1

        serial = self.ADAPTER_MSECS.dump_python(delta)
        assert isinstance(serial, float)
        assert int(serial) == 1234

    @pytest.mark.parametrize(
        'type_',
        [
            Annotated[int, TrBeforeTimedelta()],
            Annotated[bytes, TrBeforeTimedelta()],
        ],
    )
    def test_with_error(self, type_):
        with pytest.raises(TypeError, match='source type is not a timedelta'):
            TypeAdapter(type_)


class TestTrEd25519PrivateKey:
    KEY_TYPE = Ed25519PrivateKey
    ADAPTER_RAW = TypeAdapter(Annotated[Ed25519PrivateKey, TrEd25519PrivateKey()])
    ADAPTER_ENC = TypeAdapter(
        Annotated[
            Ed25519PrivateKey,
            TrEd25519PrivateKey(expanded=False),
            EncodedBytes(encoder=Base64Encoder),
        ],
    )
    ADAPTER_EXP = TypeAdapter(
        Annotated[
            Ed25519PrivateKey,
            TrEd25519PrivateKey(expanded=True),
            EncodedBytes(encoder=Base64Encoder),
        ]
    )
    TEST_KEY = 'czJbjz9SLJqx6DVIRe1cWTSWXM4UeYiRNTnAPYGDlMU='
    EXPANDED = (
        '0EqCqB0L1FnKrZwu6ovSwCD3gEfWVxVAAlJiToTI3Ea6fC2IxwcKJt4MCEuc9oQo'
        'kYK+HdXtbc3jIvySyLaNMg'
    )
    EXPECTED = TEST_KEY.rstrip('=')

    @pytest.mark.parametrize(
        'raw',
        [
            EXPECTED,
            TEST_KEY,
            b64decode(TEST_KEY),
            Ed25519PrivateKey.from_private_bytes(b64decode(TEST_KEY)),
        ],
    )
    def test_decode_encode(self, raw):
        key = self.ADAPTER_ENC.validate_python(raw)
        assert isinstance(key, self.KEY_TYPE)

        serial = self.ADAPTER_ENC.dump_python(key)
        assert serial == self.EXPECTED

    def test_expanded_key_parse(self):
        with pytest.raises(ValidationError, match='An Ed25519 private key is 32 bytes long'):
            self.ADAPTER_EXP.validate_python(self.EXPANDED)

    def test_expanded_key_serialize(self):
        key = Ed25519PrivateKey.from_private_bytes(b64decode(self.TEST_KEY))
        ser = self.ADAPTER_EXP.dump_python(key)
        assert ser == self.EXPANDED

    def test_using_raw_bytes(self):
        raw = b64decode(self.TEST_KEY)
        key = self.ADAPTER_RAW.validate_python(raw)
        assert self.ADAPTER_RAW.dump_python(key) == raw


class TestTrEd25519PublicKey:
    KEY_TYPE = Ed25519PublicKey
    ADAPTER_RAW = TypeAdapter(Annotated[Ed25519PublicKey, TrEd25519PublicKey()])
    ADAPTER_ENC = TypeAdapter(
        Annotated[
            Ed25519PublicKey,
            TrEd25519PublicKey(),
            EncodedBytes(encoder=Base32Encoder),
        ],
    )
    TEST_KEY = 'LQGMCX7HKXJZ52KH2U5KABXIUTN6MGIYIVCNQQGMJBRF24QT5UOA===='
    EXPECTED = TEST_KEY.rstrip('=')

    @pytest.mark.parametrize(
        'raw',
        [
            TEST_KEY,
            EXPECTED,
            b32decode(TEST_KEY),
            Ed25519PublicKey.from_public_bytes(b32decode(TEST_KEY)),
        ],
    )
    def test_decode_encode(self, raw):
        key = self.ADAPTER_ENC.validate_python(raw)
        assert isinstance(key, self.KEY_TYPE)

        serial = self.ADAPTER_ENC.dump_python(key)
        assert serial == self.EXPECTED

    def test_using_raw_bytes(self):
        raw = b32decode(self.TEST_KEY)
        key = self.ADAPTER_RAW.validate_python(raw)
        assert self.ADAPTER_RAW.dump_python(key) == raw


class TestTrRSAPrivateKey:
    KEY_TYPE = RSAPrivateKey
    ADAPTER_RAW = TypeAdapter(Annotated[RSAPrivateKey, TrRSAPrivateKey()])
    ADAPTER_ENC = TypeAdapter(
        Annotated[
            RSAPrivateKey,
            TrRSAPrivateKey(),
            EncodedBytes(encoder=Base64Encoder),
        ],
    )
    # This key was generated using ADD_ONION NEW:RSA1024 on Tor 0.4.5.9
    TEST_STRING = (
        'MIICWwIBAAKBgQCkRs9KZFlgLhts7ASor0dUb9RlpKlTNJarif+n041xhCEqGxEt'
        'U23G9SOsYS8L6cIWWw53YJkAzzHc/2qVB8Yxv9dLPT/YwDswHLgUbD6XP1TRtu0i'
        '3jLURDv9tc3tWn7esoMpezs5TPgvwLOxAfTXN332GnDiek+bwvcMxHPBSwIDAQAB'
        'AoGAOWJW1Mi7A8L3Z5QGiJo504AA9MSRNXSAUUmyWYCnvwiFwTyVQn0LMt286WFF'
        'Wub8Gm0SX5cJu2OlKmq6Y3bEv1raLw/MPWEViV8bI9AqttUCDAHGlrFhS3anq8it'
        'ZZpuillaffz+AUX4Od0HUbFaGSnOG89CbAlyq2qTYfyiqTkCQQDO4qGcsH9P/ZNi'
        'kwM+MK+MvkdaKx+dME+lYMbcU/BRfKzTH9ie5y1J6SaA+rMnvMN8uYP/C8D39zwk'
        'J2dnylM/AkEAy0anQbfZqf++6ng7YQpZW/V6uFXlPZaQeXjK4+KnMHslj9lSDUBr'
        'MoLxtP2zIonQqIO52XupBQbVIlGuMiZq9QJAILURqcz5g7LqLyZg198okc6vRyEU'
        'MWym2tVu+vxGPQvB4urg+1Y/AbVbgf6gfkLIgRpvNM4t5sXueyTDo1QITwJASG2B'
        'NMpEFO1Z4gM67QWZ90kNE9cPGhWmnpFqgS4F8iE+rfV55dzZFSNQ6fMnO5wtK43b'
        'z2DfRTo9AMBnt9i2bQJAc3vZmNbslZsiUXp20jmNWzWov+q8Yk5ZP9AZFWJeLCKt'
        'VgSw4x1yEH1Y+pI+3V4arQxaoQ00iNbK3Ticdg0DIw=='
    )
    TEST_BYTES = b64decode(TEST_STRING)
    EXPECTED = TEST_STRING.rstrip('=')

    @pytest.mark.parametrize(
        'raw',
        [
            TEST_STRING,
            TEST_BYTES,
            EXPECTED,
            TrRSAPrivateKey().from_bytes(TEST_BYTES),
        ],
    )
    def test_decode_encode(self, raw):
        key = self.ADAPTER_ENC.validate_python(raw)
        assert isinstance(key, self.KEY_TYPE)

        serial = self.ADAPTER_ENC.dump_python(key)
        assert serial == self.EXPECTED

    def test_using_raw_bytes(self):
        key = self.ADAPTER_RAW.validate_python(self.TEST_BYTES)
        assert self.ADAPTER_RAW.dump_python(key) == self.TEST_BYTES

    def test_using_ed25519_private_key(self):
        key = 'MC4CAQAwBQYDK2VwBCIEIOt6WDTJqbRry3WJ30ZNynCPwLaFQ114NaYr3spHpvVi'
        with pytest.raises(TypeError, match='Loaded key is not a valid RSA private key'):
            self.ADAPTER_ENC.validate_python(key)


class TestTrRSAPublicKey:
    KEY_TYPE = RSAPublicKey
    ADAPTER_RAW = TypeAdapter(Annotated[RSAPublicKey, TrRSAPublicKey()])
    ADAPTER_ENC = TypeAdapter(
        Annotated[
            RSAPublicKey,
            TrRSAPublicKey(),
            EncodedBytes(encoder=Base64Encoder),
        ],
    )
    TEST_STRING = (
        'MIGJAoGBAN87FeyffLsjGGgd6qxVLLoKWG1GXfyu4o6OrRi1a5Mv0bgRmwqfo1O5'
        'g/c7/6JqhNFYd0UNIzyGB2LBHgQJwYTcPTW//Gpn8Dfcysl0gjA4+MLU/xQ/24Vi'
        'bbkL05YLK0AabiedS3Pmm5bzy05xAdRFitK22BeXLYDuRBGrIVejAgMBAAE='
    )
    TEST_BYTES = b64decode(TEST_STRING)
    EXPECTED = TEST_STRING.rstrip('=')

    @pytest.mark.parametrize(
        'raw',
        [
            TEST_STRING,
            TEST_BYTES,
            EXPECTED,
            TrRSAPublicKey().from_bytes(TEST_BYTES),
        ],
    )
    def test_decode_encode(self, raw):
        key = self.ADAPTER_ENC.validate_python(raw)
        assert isinstance(key, self.KEY_TYPE)

        serial = self.ADAPTER_ENC.dump_python(key)
        assert serial == self.EXPECTED

    def test_using_raw_bytes(self):
        key = self.ADAPTER_RAW.validate_python(self.TEST_BYTES)
        assert self.ADAPTER_RAW.dump_python(key) == self.TEST_BYTES

    def test_using_ed25519_private_key(self):
        key = 'MCowBQYDK2VwAyEA5AcvvI6f7k6+VGVxr9KLRrjoHE9CZQGDSj5XnAXDUeY'
        with pytest.raises(TypeError, match='Loaded key is not a valid RSA public key'):
            self.ADAPTER_ENC.validate_python(key)


class TestTrX25519PrivateKey:
    KEY_TYPE = X25519PrivateKey
    ADAPTER_RAW = TypeAdapter(Annotated[X25519PrivateKey, TrX25519PrivateKey()])
    ADAPTER_ENC = TypeAdapter(
        Annotated[
            X25519PrivateKey,
            TrX25519PrivateKey(),
            EncodedBytes(encoder=Base64Encoder),
        ],
    )
    TEST_KEY = 'yPGUxgKaC5ACyEzsdANHJEJzt5DIqDRBlAFaAWWQn0o='
    EXPECTED = TEST_KEY.rstrip('=')

    @pytest.mark.parametrize(
        'raw',
        [
            EXPECTED,
            TEST_KEY,
            b64decode(TEST_KEY),
            X25519PrivateKey.from_private_bytes(b64decode(TEST_KEY)),
        ],
    )
    def test_decode_encode(self, raw):
        key = self.ADAPTER_ENC.validate_python(raw)
        assert isinstance(key, self.KEY_TYPE)

        serial = self.ADAPTER_ENC.dump_python(key)
        assert serial == self.EXPECTED

    def test_using_raw_bytes(self):
        raw = b64decode(self.TEST_KEY)
        key = self.ADAPTER_RAW.validate_python(raw)
        assert self.ADAPTER_RAW.dump_python(key) == raw

    @pytest.mark.parametrize(
        'type_',
        [
            Annotated[int, TrX25519PrivateKey()],
            Annotated[None, TrX25519PrivateKey()],
        ],
    )
    def test_usage_error_on_source_type(self, type_):
        with pytest.raises(TypeError, match='source type is not a X25519PrivateKey'):
            TypeAdapter(type_)


class TestTrX25519PublicKey:
    KEY_TYPE = X25519PublicKey
    ADAPTER_RAW = TypeAdapter(Annotated[X25519PublicKey, TrX25519PublicKey()])
    ADAPTER_ENC = TypeAdapter(
        Annotated[
            X25519PublicKey,
            TrX25519PublicKey(),
            EncodedBytes(encoder=Base32Encoder),
        ],
    )
    TEST_KEY = 'K2MLQ4S2DS4YCZXDOTOVC45LCLAKKCKN7QVAXPDMOSSYPZBGQSLA===='
    EXPECTED = TEST_KEY.rstrip('=')

    @pytest.mark.parametrize(
        'raw',
        [
            TEST_KEY,
            EXPECTED,
            b32decode(TEST_KEY),
            X25519PublicKey.from_public_bytes(b32decode(TEST_KEY)),
        ],
    )
    def test_decode_encode(self, raw):
        key = self.ADAPTER_ENC.validate_python(raw)
        assert isinstance(key, self.KEY_TYPE)

        serial = self.ADAPTER_ENC.dump_python(key)
        assert serial == self.EXPECTED

    def test_using_raw_bytes(self):
        raw = b32decode(self.TEST_KEY)
        key = self.ADAPTER_RAW.validate_python(raw)
        assert self.ADAPTER_RAW.dump_python(key) == raw


class TestTrBoolYesNo:
    ADAPTER = TypeAdapter(Annotated[bool, TrBoolYesNo()])

    @pytest.mark.parametrize(
        ('entry', 'bval', 'serial'),
        [
            ('YES', True, 'yes'),
            (True, True, 'yes'),
            ('NO', False, 'no'),
            (False, False, 'no'),
        ],
    )
    def test_parse_and_encode(self, entry, bval, serial):
        res = self.ADAPTER.validate_python(entry)
        assert res is bval
        ser = self.ADAPTER.dump_python(res)
        assert ser == serial


class TestTrCast:
    ADAPTER_AFTER = TypeAdapter(
        Annotated[
            int,
            BeforeValidator(lambda x: 2 * x),
            TrCast(float, mode='after'),
        ],
    )
    ADAPTER_BEFORE = TypeAdapter(
        Annotated[
            int,
            BeforeValidator(lambda x: 2 * x),
            TrCast(float, mode='before'),
        ],
    )

    def test_after(self):
        r = self.ADAPTER_AFTER.validate_python('12')
        assert isinstance(r, float)
        assert r == 1212.0

    def test_before(self):
        r = self.ADAPTER_BEFORE.validate_python('12')
        assert isinstance(r, int)
        assert r == 24



from __future__ import annotations

from asyncio import StreamReader

import pytest

from aiostem.exceptions import ProtocolError
from aiostem.utils import MessageData, MessageLine, messages_from_stream

# All test coroutines will be treated as marked for asyncio.
pytestmark = pytest.mark.asyncio


def serialize(lines: list[str]) -> str:
    """
    Serialize lines to build a message suitable for the stream.

    Args:
        lines: a list of lines to serialize.

    Returns:
        A serialized text message.

    """
    return '\r\n'.join(lines) + '\r\n'


def create_stream(lines: list[str]) -> StreamReader:
    """
    Create a new stream reader that was fed with initial lines.

    Args:
        lines: a list of initial lines.

    Returns:
        A stream reader already fed with the provided lines.

    """
    data = serialize(lines)
    stream = StreamReader()
    stream.feed_data(data.encode('ascii'))
    stream.feed_eof()
    return stream


class TestMessage:
    async def test_simple_message(self):
        """Check that we can parse a single simple message."""
        stream = create_stream(['650 DEBUG'])
        messages = [msg async for msg in messages_from_stream(stream)]
        assert len(messages) == 1

        message = messages[0]
        assert message.status == 650
        assert message.header == 'DEBUG'
        assert message.is_event is True
        assert message.is_success is False
        assert message.is_error is False

    async def test_simple_serialized(self):
        """Check that our simple message can be serialized back."""
        lines = ['650 DEBUG']
        stream = create_stream(['650 DEBUG'])
        messages = [msg async for msg in messages_from_stream(stream)]
        text = serialize(lines)
        assert messages[0].serialize() == text

    async def test_multiple_simple_messages(self):
        stream = create_stream(['650 DEBUG', '250 OK', '250 LOG'])
        messages = [msg async for msg in messages_from_stream(stream)]
        assert len(messages) == 3

        assert messages[0].status == 650
        assert messages[0].header == 'DEBUG'
        assert messages[1].status == 250
        assert messages[1].header == 'OK'
        assert messages[2].status == 250
        assert messages[2].header == 'LOG'

    async def test_with_sub_messages(self):
        lines = [
            '650-key=value',
            '650+multi=',
            '..DOT-VALUE',
            'SECOND-LINE',
            '.',
            '650 DEBUG',
        ]
        stream = create_stream(lines)
        messages = [msg async for msg in messages_from_stream(stream)]
        assert len(messages) == 1

        message = messages[0]
        assert message.status == 650
        assert message.header == 'DEBUG'
        assert len(message.items) == 2
        assert isinstance(message.items[0], MessageLine)
        assert isinstance(message.items[1], MessageData)

        serialized = serialize(lines)
        assert serialized == message.serialize()

    @pytest.mark.parametrize(
        ('line', 'error'),
        [
            ('65', 'Received line is too short'),
            ('65X ', 'Unable to parse status code'),
            ('650_', 'Unrecognized separator'),
        ],
    )
    async def test_protocol_errors(self, line, error):
        stream = create_stream([line])
        iterator = messages_from_stream(stream)
        with pytest.raises(ProtocolError, match=error):
            await iterator.__anext__()

    async def test_basic_keyword(self):
        stream = create_stream(
            [
                '250-PROTOCOLINFO 1',
                '250-AUTH METHODS=HASHEDPASSWORD',
                '250-VERSION Tor="0.4.8.12"',
                '250 OK',
            ]
        )
        messages = [msg async for msg in messages_from_stream(stream)]
        assert len(messages) == 1
        message = messages[0]
        assert message.keyword == 'OK'
        assert message.items[0].keyword == 'PROTOCOLINFO'
        assert message.items[1].keyword == 'AUTH'
        assert message.items[2].keyword == 'VERSION'

    async def test_event_keyword_single(self):
        """Check that a keyword on an event works well."""
        stream = create_stream(['650 CIRC KEY=VALUE'])
        messages = [msg async for msg in messages_from_stream(stream)]
        assert len(messages) == 1
        assert messages[0].keyword == 'CIRC'

    async def test_event_keyword_multi(self):
        """Check that a keyword on a multi-message event in on the first item."""
        stream = create_stream(['650-CIRC KEY=VALUE', '650 OK'])
        messages = [msg async for msg in messages_from_stream(stream)]
        assert len(messages) == 1
        assert messages[0].keyword == 'CIRC'



from __future__ import annotations

import logging
from collections.abc import Sequence

import pytest
from pydantic import TypeAdapter

from aiostem.event import EventBaseNetworkStatus
from aiostem.structures import RouterStatus

# All test coroutines will be treated as marked for asyncio.
pytestmark = pytest.mark.asyncio


class TestLiveData:
    """Test with live data issued by Tor."""

    @pytest.mark.timeout(10)
    async def test_network_status(self, controller, caplog):
        """
        Check the network statuses of onion routers.

        This means that we are (still) able to parse network statuses produced
        by Tor and that their format does not contain unknown directives.

        """
        reply = await controller.get_info('ns/all')
        reply.raise_for_status()

        adapter = TypeAdapter(Sequence[RouterStatus])
        with caplog.at_level(logging.WARNING, logger='aiostem'):
            data = EventBaseNetworkStatus.parse_router_statuses(reply['ns/all'])
            results = adapter.validate_python(data)

        assert len(caplog.records) == 0
        assert len(results) > 0



from __future__ import annotations

import hashlib
import typing
from abc import ABC, abstractmethod
from collections.abc import (
    Collection,
    Mapping,
    MutableSequence,  # noqa: F401
    Sequence,
    Set as AbstractSet,
)
from dataclasses import dataclass, field
from datetime import UTC, datetime, timedelta, tzinfo
from typing import TYPE_CHECKING, Any, ClassVar, Generic, Literal, TypeVar

from cryptography.hazmat.primitives.asymmetric.ed25519 import (
    Ed25519PrivateKey,
    Ed25519PublicKey,
)
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey
from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PrivateKey, X25519PublicKey
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    NoEncryption,
    PrivateFormat,
    PublicFormat,
    load_der_private_key,
    load_der_public_key,
)
from pydantic import ConfigDict
from pydantic_core import core_schema
from pydantic_core.core_schema import CoreSchema, WhenUsed

if TYPE_CHECKING:
    from pydantic import GetCoreSchemaHandler
    from pydantic_core.core_schema import SerializerFunctionWrapHandler


@dataclass(frozen=True, slots=True)
class TrAfterAsTimezone:
    """Post-validator that enforces a timezone."""

    #: Timezone to map this date to.
    timezone: tzinfo = UTC

    def __get_pydantic_core_schema__(
        self,
        source: type[Any],
        handler: GetCoreSchemaHandler,
    ) -> CoreSchema:
        """Declare a validator to add or change the timezone."""
        if not issubclass(source, datetime):
            msg = f"source type is not a datetime, got '{source.__name__}'"
            raise TypeError(msg)

        return core_schema.chain_schema(
            steps=[
                core_schema.datetime_schema(),
                core_schema.no_info_after_validator_function(
                    function=self.from_value,
                    schema=core_schema.datetime_schema(),
                ),
            ],
        )

    def from_value(self, value: datetime) -> datetime:
        """
        Apply the timezone of change the offset.

        Args:
            value: The original datetime to change.

        Returns:
            A new datetime with the proper timezone applied.

        """
        if value.tzinfo is None:
            return value.replace(tzinfo=self.timezone)
        return value.astimezone(self.timezone)


@dataclass(frozen=True, slots=True)
class TrCast:
    """Pre-validator that converts to the target type."""

    #: Type we want to cast this to!
    target: type[Any]

    #: Whether to apply this transformation before or after the main handler.
    mode: Literal['before', 'after'] = 'before'

    def __get_pydantic_core_schema__(
        self,
        source: type[Any],
        handler: GetCoreSchemaHandler,
    ) -> CoreSchema:
        """Declare schema and validator to cast to the provided type."""
        schema: CoreSchema
        source_schema = handler(source)
        target_schema = handler.generate_schema(self.target)

        match self.mode:
            case 'before':
                schema = core_schema.no_info_before_validator_function(
                    function=self._pydantic_validator,
                    schema=core_schema.union_schema(
                        choices=[
                            core_schema.chain_schema(
                                steps=[
                                    target_schema,
                                    source_schema,
                                ],
                            ),
                            source_schema,
                        ]
                    ),
                )

            case 'after':  # pragma: no branch
                schema = core_schema.no_info_after_validator_function(
                    function=self._pydantic_validator,
                    schema=core_schema.union_schema(
                        choices=[
                            core_schema.chain_schema(
                                steps=[
                                    source_schema,
                                    target_schema,
                                ],
                            ),
                            source_schema,
                        ]
                    ),
                )

        return schema

    def _pydantic_validator(self, value: Any) -> Any:
        """Do not validate, simply pass the value."""
        return value


@dataclass(frozen=True, slots=True)
class TrBeforeSetToNone:
    """Pre-validator that sets a value to :obj:`None`."""

    #: List of values mapped to None.
    values: AbstractSet[Any] = field(default_factory=frozenset)

    def __get_pydantic_core_schema__(
        self,
        source: type[Any],
        handler: GetCoreSchemaHandler,
    ) -> CoreSchema:
        """Declare schema and validator to replace values to None."""
        return core_schema.no_info_before_validator_function(
            function=self.from_value,
            schema=handler(source),
        )

    def from_value(self, value: Any) -> Any:
        """
        Set the return value to :obj:`None` when applicable.

        Args:
            value: The value to check against :attr:`values`.

        Returns:
            The same value of :obj:`None` when matching any value in :attr:`values`.

        """
        if value in self.values:
            return None
        return value


@dataclass(frozen=True, slots=True)
class TrBeforeStringSplit:
    """Deserialize sequences from/to strings."""

    #: Base pydantic configuration to apply when serializing.
    model_config: ClassVar[ConfigDict | None] = None

    #: Maximum number of string split.
    maxsplit: int = -1

    #: How to split this string sequence.
    separator: str = ','

    #: Optional list of keys when converted to a dictionary.
    dict_keys: Sequence[str] | None = None

    #: When serialization is supposed to be used.
    when_used: WhenUsed = 'always'

    def __get_pydantic_core_schema__(
        self,
        source: type[Any],
        handler: GetCoreSchemaHandler,
    ) -> CoreSchema:
        """Tell the core schema and how to validate the whole thing."""
        if self.dict_keys is None:
            # Check that we have a valid collection of something else.
            origin = typing.get_origin(source) or source
            if not isinstance(origin, type) or not issubclass(origin, Collection):
                msg = f"source type is not a collection, got '{source.__name__}'"
                raise TypeError(msg)

        return core_schema.union_schema(
            choices=[
                handler(source),
                core_schema.chain_schema(
                    steps=[
                        core_schema.str_schema(),
                        core_schema.no_info_before_validator_function(
                            function=self._pydantic_validator,
                            schema=handler(source),
                        ),
                    ],
                ),
            ],
            serialization=core_schema.wrap_serializer_function_ser_schema(
                function=self._pydantic_serializer,
                schema=handler(source),
                return_schema=core_schema.str_schema(),
                when_used=self.when_used,
            ),
        )

    def _pydantic_validator(self, value: str) -> Sequence[str] | Mapping[str, str]:
        """Parse the input string and convert it to a list or a dictionary."""
        items = value.split(self.separator, maxsplit=self.maxsplit)
        if self.dict_keys is not None:
            return dict(zip(self.dict_keys, items, strict=False))
        return items

    def _pydantic_serializer(
        self,
        value: Any,
        serializer: SerializerFunctionWrapHandler,
    ) -> str:
        """Tells how we serialize this collection for JSON."""
        values = []  # type: MutableSequence[str]
        parts = serializer(value)

        if isinstance(parts, Mapping) and isinstance(self.dict_keys, Sequence):
            for key in self.dict_keys:
                values.append(parts[key])
        else:
            values.extend(parts)
        return self.separator.join(map(str, values))


@dataclass(frozen=True, slots=True)
class TrBeforeTimedelta:
    """Pre-validator that gets a timedelta from an int or float."""

    #: Unit used to serialize and deserialize.
    unit: Literal['milliseconds', 'seconds', 'minutes'] = 'seconds'

    #: Whether to serialize / deserialize to a float number.
    is_float: bool = True

    def __get_pydantic_core_schema__(
        self,
        source: type[Any],
        handler: GetCoreSchemaHandler,
    ) -> CoreSchema:
        """Set a custom validator used to transform seconds in a timedelta."""
        if not issubclass(source, timedelta):
            msg = f"source type is not a timedelta, got '{source.__name__}'"
            raise TypeError(msg)

        if self.is_float:
            source_schema = core_schema.float_schema()  # type: CoreSchema
        else:
            source_schema = core_schema.int_schema()

        return core_schema.union_schema(
            choices=[
                handler(source),
                core_schema.chain_schema(
                    steps=[
                        source_schema,
                        core_schema.no_info_before_validator_function(
                            self.from_number,
                            handler(source),
                        ),
                    ],
                ),
            ],
            serialization=core_schema.plain_serializer_function_ser_schema(
                function=self.to_number,
                return_schema=source_schema,
            ),
        )

    def from_number(self, value: float) -> timedelta:
        """
        Parse the input value as an integer or float timedelta.

        Args:
            value: The input value we want to create a timedelta from.

        Returns:
            A simple timedelta built from the provided value.

        """
        match self.unit:
            case 'milliseconds':
                value = value / 1000.0
            case 'minutes':
                value = value * 60.0
        return timedelta(seconds=value)

    def to_number(self, delta: timedelta) -> float | int:
        """
        Convert the timedelta value to a float or int.

        Args:
            delta: The timedelta value we want to serialize.

        Returns:
            A float or integer value.

        """
        value = delta.total_seconds()

        match self.unit:
            case 'milliseconds':
                value = 1000 * value
            case 'minutes':
                value = value / 60.0

        if not self.is_float:
            return int(value)
        return value


KeyType = TypeVar('KeyType')


class TrGenericKey(ABC, Generic[KeyType]):
    """Transform bytes in a key."""

    def __get_pydantic_core_schema__(
        self,
        source: type[Any],
        handler: GetCoreSchemaHandler,
    ) -> CoreSchema:
        """Declare schema and validator for any kind of keys."""
        key_type = self._get_type()

        if not issubclass(source, key_type):
            msg = f"source type is not a {key_type.__name__}, got '{source.__name__}'"
            raise TypeError(msg)

        source_schema = core_schema.bytes_schema(strict=True)
        return core_schema.union_schema(
            choices=[
                core_schema.is_instance_schema(key_type),
                core_schema.no_info_after_validator_function(
                    function=self.from_bytes,
                    schema=source_schema,
                ),
            ],
            serialization=core_schema.plain_serializer_function_ser_schema(
                function=self.to_bytes,
                return_schema=source_schema,
            ),
        )

    @abstractmethod
    def _get_type(self) -> type[KeyType]:
        """Get the key type used by this generic class."""

    @abstractmethod
    def from_bytes(self, data: bytes) -> KeyType:
        """Build a key from the provided bytes."""

    @abstractmethod
    def to_bytes(self, key: KeyType) -> bytes:
        """Serialize a key to its corresponding bytes."""


@dataclass(frozen=True, slots=True)
class TrEd25519PrivateKey(TrGenericKey[Ed25519PrivateKey]):
    """
    Transform bytes into an ed25519 private key.

    Note:
        Tor's implementation of Ed25519 is donna-ed25519 and uses the expanded
        form as the private key (64 bytes). Unfortunately ``cryptography`` does
        not provide such interface, which means that we are left with this
        implementation.

    """

    #: Whether to generate the expanded form while serializing.
    #:
    #: Note:
    #:     This makes parsing impossible...
    expanded: bool = False

    def _get_type(self) -> type[Ed25519PrivateKey]:
        """Get the key type used by this generic class."""
        return Ed25519PrivateKey

    def from_bytes(self, data: bytes) -> Ed25519PrivateKey:
        """
        Build an ed25519 private key out of the provided bytes.

        Args:
            data: a 32 bytes seed representing a secret key.

        Returns:
            An instance of an ed25519 private key.

        """
        return Ed25519PrivateKey.from_private_bytes(data)

    def to_bytes(self, key: Ed25519PrivateKey) -> bytes:
        """
        Serialize the provided private key to bytes.

        See Also:
            - :meth:`to_expanded_bytes`
            - :meth:`to_seed_bytes`

        Returns:
            32 or 64 bytes corresponding to the private key.

        """
        if self.expanded:
            return self.to_expanded_bytes(key)
        return self.to_seed_bytes(key)

    def to_expanded_bytes(self, key: Ed25519PrivateKey) -> bytes:
        """
        Serialize to the expanded form.

        Returns:
            64 bytes corresponding to the expanded private key.

        """
        seed = self.to_seed_bytes(key)
        extsk = list(hashlib.sha512(seed).digest())
        extsk[0] &= 248
        extsk[31] &= 127
        extsk[31] |= 64
        return bytes(extsk)

    def to_seed_bytes(self, key: Ed25519PrivateKey) -> bytes:
        """
        Serialize the seed bytes, which is the default behavior.

        Returns:
            32 bytes corresponding to the private key.

        """
        return key.private_bytes_raw()


@dataclass(frozen=True, slots=True)
class TrEd25519PublicKey(TrGenericKey[Ed25519PublicKey]):
    """Transform bytes into a ed25519 public key."""

    def _get_type(self) -> type[Ed25519PublicKey]:
        """Get the key type used by this generic class."""
        return Ed25519PublicKey

    def from_bytes(self, data: bytes) -> Ed25519PublicKey:
        """
        Build an ed25519 public key out of the provided bytes.

        Returns:
            An instance of an ed25519 public key.

        """
        return Ed25519PublicKey.from_public_bytes(data)

    def to_bytes(self, key: Ed25519PublicKey) -> bytes:
        """
        Serialize the provided public key to bytes.

        Returns:
            32 bytes corresponding to the public key.

        """
        return key.public_bytes_raw()


@dataclass(frozen=True, slots=True)
class TrRSAPrivateKey(TrGenericKey[RSAPrivateKey]):
    """Transform bytes into a RSA private key."""

    #: Encoding format for the public RSA key.
    encoding: Encoding = Encoding.DER

    #: Key format used while serializing.
    format: PrivateFormat = PrivateFormat.TraditionalOpenSSL

    def _get_type(self) -> type[RSAPrivateKey]:
        """Get the key type used by this generic class."""
        return RSAPrivateKey

    def from_bytes(self, data: bytes) -> RSAPrivateKey:
        """
        Build a RSA private key out of the provided bytes.

        Returns:
            An instance of a RSA private key.

        """
        key = load_der_private_key(data, password=None)
        if not isinstance(key, RSAPrivateKey):
            msg = 'Loaded key is not a valid RSA private key.'
            raise TypeError(msg)
        return key

    def to_bytes(self, key: RSAPrivateKey) -> bytes:
        """Serialize the provided private key to bytes."""
        return key.private_bytes(self.encoding, self.format, NoEncryption())


@dataclass(frozen=True, slots=True)
class TrRSAPublicKey(TrGenericKey[RSAPublicKey]):
    """Transform bytes into a RSA public key."""

    #: Encoding format for the public RSA key.
    encoding: Encoding = Encoding.DER

    #: Key format used while serializing.
    format: PublicFormat = PublicFormat.PKCS1

    def _get_type(self) -> type[RSAPublicKey]:
        """Get the key type used by this generic class."""
        return RSAPublicKey

    def from_bytes(self, data: bytes) -> RSAPublicKey:
        """
        Build a RSA public key out of the provided bytes.

        Returns:
            An instance of a RSA public key.

        """
        key = load_der_public_key(data)
        if not isinstance(key, RSAPublicKey):
            msg = f'Loaded key is not a valid RSA public key: got {type(key)}.'
            raise TypeError(msg)
        return key

    def to_bytes(self, key: RSAPublicKey) -> bytes:
        """Serialize the provided public key to bytes."""
        return key.public_bytes(self.encoding, self.format)


@dataclass(frozen=True, slots=True)
class TrX25519PrivateKey(TrGenericKey[X25519PrivateKey]):
    """Transform bytes into a X25519 private key."""

    def _get_type(self) -> type[X25519PrivateKey]:
        """Get the key type used by this generic class."""
        return X25519PrivateKey

    def from_bytes(self, data: bytes) -> X25519PrivateKey:
        """
        Build a X25519 private key out of the provided bytes.

        Returns:
            An instance of a X25519 private key.

        """
        return X25519PrivateKey.from_private_bytes(data)

    def to_bytes(self, key: X25519PrivateKey) -> bytes:
        """
        Serialize the provided private key to bytes.

        Returns:
            32 bytes corresponding to the private key.

        """
        return key.private_bytes_raw()


@dataclass(frozen=True, slots=True)
class TrX25519PublicKey(TrGenericKey[X25519PublicKey]):
    """Transform bytes into a X25519 public key."""

    def _get_type(self) -> type[X25519PublicKey]:
        """Get the key type used by this generic class."""
        return X25519PublicKey

    def from_bytes(self, data: bytes) -> X25519PublicKey:
        """
        Build a X25519 public key out of the provided bytes.

        Returns:
            An instance of a X25519 public key.

        """
        return X25519PublicKey.from_public_bytes(data)

    def to_bytes(self, key: X25519PublicKey) -> bytes:
        """
        Serialize the provided public key to bytes.

        Returns:
            32 bytes corresponding to the public key.

        """
        return key.public_bytes_raw()


@dataclass(frozen=True, slots=True)
class TrBoolYesNo:
    """Transform yes/no to and from a boolean."""

    #: What we use when this value is :obj:`True`.
    true: str = 'yes'
    #: What we use when this value is :obj:`False`.
    false: str = 'no'

    def __get_pydantic_core_schema__(
        self,
        source: type[Any],
        handler: GetCoreSchemaHandler,
    ) -> CoreSchema:
        """Serialize and deserialize from/to a string."""
        return core_schema.union_schema(
            choices=[
                core_schema.bool_schema(),
            ],
            serialization=core_schema.plain_serializer_function_ser_schema(
                function=self.to_string,
                return_schema=core_schema.str_schema(strict=True),
            ),
        )

    def to_string(self, value: bool) -> str:
        """Serialize this value to a string."""
        return self.true if value else self.false



from __future__ import annotations

from datetime import datetime, timedelta
from ipaddress import IPv4Address, IPv6Address
from typing import Annotated, Generic, TypeAlias, TypeVar, Union

from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PublicKey
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PrivateKey, X25519PublicKey
from pydantic import BaseModel, Field

from .utils import (
    Base16Encoder,
    Base32Encoder,
    Base64Encoder,
    EncodedBytes,
    TrAfterAsTimezone,
    TrBeforeTimedelta,
    TrBoolYesNo,
    TrEd25519PublicKey,
    TrRSAPublicKey,
    TrX25519PrivateKey,
    TrX25519PublicKey,
)

#: Any boundary of a range structure (int, float, etc...).
RangeVal = TypeVar('RangeVal')


# Generic models do not work well with dataclasses even on recent pydantic :(.
class GenericRange(BaseModel, Generic[RangeVal]):
    """Any kind of numeric range."""

    #: Minimum value in the range (inclusive).
    min: RangeVal
    #: Maximum value in the range (inclusive).
    max: RangeVal


#: Any IP address, either IPv4 or IPv6.
AnyAddress: TypeAlias = Union[IPv4Address | IPv6Address]  # noqa: UP007

#: Any host, either by IP address or hostname.
AnyHost: TypeAlias = Annotated[
    IPv4Address | IPv6Address | str,
    Field(union_mode='left_to_right'),
]
#: Any TCP or UDP port.
AnyPort: TypeAlias = Annotated[int, Field(gt=0, lt=65536)]

#: Bytes that are hex encoded.
Base16Bytes: TypeAlias = Annotated[bytes, EncodedBytes(encoder=Base16Encoder)]

#: Bytes that are base32 encoded.
Base32Bytes: TypeAlias = Annotated[bytes, EncodedBytes(encoder=Base32Encoder)]

#: Bytes that are base64 encoded.
Base64Bytes: TypeAlias = Annotated[bytes, EncodedBytes(encoder=Base64Encoder)]

#: A boolean value serialized as a yes/no string.
BoolYesNo: TypeAlias = Annotated[bool, TrBoolYesNo()]

#: Base64 encoded bytes parsed as an ed25519 public key.
Ed25519PublicKeyBase64 = Annotated[
    Ed25519PublicKey,
    TrEd25519PublicKey(),
    EncodedBytes(encoder=Base64Encoder),
]

#: A datetime that always puts or convert to UTC.
DatetimeUTC: TypeAlias = Annotated[datetime, TrAfterAsTimezone()]

#: Base64 encoded bytes parsed as a public RSA key.
RSAPublicKeyBase64: TypeAlias = Annotated[
    RSAPublicKey,
    TrRSAPublicKey(),
    EncodedBytes(encoder=Base64Encoder),
]

#: A :class:`~datetime.timedelta` parsed from an integer value in milliseconds.
TimedeltaMilliseconds: TypeAlias = Annotated[
    timedelta,
    TrBeforeTimedelta(unit='milliseconds'),
]
#: A :class:`~datetime.timedelta` parsed from an integer value in minutes.
TimedeltaMinutesInt: TypeAlias = Annotated[
    timedelta,
    TrBeforeTimedelta(unit='minutes', is_float=False),
]
#: A :class:`~datetime.timedelta` parsed from an integer value in seconds.
TimedeltaSeconds: TypeAlias = Annotated[
    timedelta,
    TrBeforeTimedelta(unit='seconds'),
]

#: Base32 encoded bytes parsed as a public x25519 key.
X25519PublicKeyBase32: TypeAlias = Annotated[
    X25519PublicKey,
    TrX25519PublicKey(),
    EncodedBytes(encoder=Base32Encoder),
]

#: Base64 encoded bytes parsed as a public x25519 key.
X25519PublicKeyBase64: TypeAlias = Annotated[
    X25519PublicKey,
    TrX25519PublicKey(),
    EncodedBytes(encoder=Base64Encoder),
]

#: Base64 encoded bytes parsed as a private x25519 key.
X25519PrivateKeyBase64: TypeAlias = Annotated[
    X25519PrivateKey,
    TrX25519PrivateKey(),
    EncodedBytes(encoder=Base64Encoder),
]



from __future__ import annotations

import aiostem


class TestVersion:
    def test_version_attribute_is_present(self):
        assert hasattr(aiostem, '__version__')

    def test_version_attribute_is_a_string(self):
        assert isinstance(aiostem.__version__, str)
        assert aiostem.version == aiostem.__version__



from __future__ import annotations

from .argument import ArgumentKeyword, ArgumentString, QuoteStyle
from .encoding import (
    Base16Encoder,
    Base32Encoder,
    Base64Encoder,
    EncodedBase,
    EncodedBytes,
    EncoderProtocol,
)
from .message import BaseMessage, Message, MessageData, MessageLine, messages_from_stream
from .syntax import ReplySyntax, ReplySyntaxFlag
from .transformers import (
    TrAfterAsTimezone,
    TrBeforeSetToNone,
    TrBeforeStringSplit,
    TrBeforeTimedelta,
    TrBoolYesNo,
    TrCast,
    TrEd25519PrivateKey,
    TrEd25519PublicKey,
    TrRSAPrivateKey,
    TrRSAPublicKey,
    TrX25519PrivateKey,
    TrX25519PublicKey,
)

__all__ = [
    'ArgumentKeyword',
    'ArgumentString',
    'Base16Encoder',
    'Base32Encoder',
    'Base64Encoder',
    'BaseMessage',
    'EncodedBase',
    'EncodedBytes',
    'EncoderProtocol',
    'Message',
    'MessageData',
    'MessageLine',
    'QuoteStyle',
    'ReplySyntax',
    'ReplySyntaxFlag',
    'TrAfterAsTimezone',
    'TrBeforeSetToNone',
    'TrBeforeStringSplit',
    'TrBeforeTimedelta',
    'TrBoolYesNo',
    'TrCast',
    'TrEd25519PrivateKey',
    'TrEd25519PublicKey',
    'TrRSAPrivateKey',
    'TrRSAPublicKey',
    'TrX25519PrivateKey',
    'TrX25519PublicKey',
    'messages_from_stream',
]



from __future__ import annotations

import asyncio
from abc import ABC, abstractmethod

#: Default path when connecting to a controller through an UNIX socket.
DEFAULT_CONTROL_PATH: str = '/var/run/tor/control'
#: Default host value when connecting to a TCP controller.
DEFAULT_CONTROL_HOST: str = '127.0.0.1'
#: Default port value when connecting to a TCP controller.
DEFAULT_CONTROL_PORT: int = 9051


class ControlConnector(ABC):
    """
    Base class for all connector types used by the controller.

    These are simply helper classes providing a pair of reader and writer needed
    to perform actions on the target control port.

    """

    @abstractmethod
    async def connect(self) -> tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        """
        Open an asynchronous connection to the target control port.

        Returns:
            A tuple forming a full-duplex asyncio stream.

        """


class ControlConnectorPort(ControlConnector):
    """Tor connector using a local or report TCP port."""

    def __init__(
        self,
        host: str = DEFAULT_CONTROL_HOST,
        port: int = DEFAULT_CONTROL_PORT,
    ) -> None:
        """
        Create a controller connector using a TCP host and port.

        Hint:
            Use :meth:`.Controller.from_port` for an automated use of this class.

        Args:
            host: IP address or hostname to the control host.
            port: TCP port to connect to.

        """
        self._host = host
        self._port = port

    @property
    def host(self) -> str:
        """IP address or host name to Tor's control port."""
        return self._host

    @property
    def port(self) -> int:
        """TCP port used to reach the control port."""
        return self._port

    async def connect(self) -> tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        """
        Open an asynchronous connection to the target's TCP port.

        Returns:
            A tuple forming a full-duplex asyncio stream.

        """
        return await asyncio.open_connection(self.host, self.port)


class ControlConnectorPath(ControlConnector):
    """Tor connector using a local unix socket."""

    def __init__(self, path: str = DEFAULT_CONTROL_PATH) -> None:
        """
        Create a controller connector using a local unix socket.

        Hint:
            Use :meth:`.Controller.from_path` for an automated use of this class.

        Args:
            path: Path to the unix socket on the local file system.

        """
        self._path = path

    @property
    def path(self) -> str:
        """Get the path to the local unix socket to Tor's control port."""
        return self._path

    async def connect(self) -> tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        """
        Open an asynchronous connection to the target unix socket.

        Returns:
            A tuple forming a full-duplex asyncio stream.

        """
        return await asyncio.open_unix_connection(self.path)



from __future__ import annotations

import asyncio
import logging
from collections.abc import Awaitable, Callable
from contextlib import AsyncExitStack, suppress
from typing import TYPE_CHECKING, TypeAlias, cast

from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey

from .command import (
    Command,
    CommandAddOnion,
    CommandAttachStream,
    CommandAuthChallenge,
    CommandAuthenticate,
    CommandCloseCircuit,
    CommandCloseStream,
    CommandDelOnion,
    CommandDropGuards,
    CommandDropOwnership,
    CommandDropTimeouts,
    CommandExtendCircuit,
    CommandGetConf,
    CommandGetInfo,
    CommandHsFetch,
    CommandHsPost,
    CommandLoadConf,
    CommandMapAddress,
    CommandOnionClientAuthAdd,
    CommandOnionClientAuthRemove,
    CommandOnionClientAuthView,
    CommandPostDescriptor,
    CommandProtocolInfo,
    CommandQuit,
    CommandRedirectStream,
    CommandResetConf,
    CommandResolve,
    CommandSaveConf,
    CommandSetCircuitPurpose,
    CommandSetConf,
    CommandSetEvents,
    CommandSignal,
    CommandTakeOwnership,
)
from .connector import (
    DEFAULT_CONTROL_HOST,
    DEFAULT_CONTROL_PATH,
    DEFAULT_CONTROL_PORT,
    ControlConnector,
    ControlConnectorPath,
    ControlConnectorPort,
)
from .event import Event, EventWord, EventWordInternal, event_from_message
from .exceptions import CommandError, ControllerError
from .reply import (
    ReplyAddOnion,
    ReplyAttachStream,
    ReplyAuthChallenge,
    ReplyAuthenticate,
    ReplyCloseCircuit,
    ReplyCloseStream,
    ReplyDelOnion,
    ReplyDropGuards,
    ReplyDropOwnership,
    ReplyDropTimeouts,
    ReplyExtendCircuit,
    ReplyGetConf,
    ReplyGetInfo,
    ReplyHsFetch,
    ReplyHsPost,
    ReplyLoadConf,
    ReplyMapAddress,
    ReplyOnionClientAuthAdd,
    ReplyOnionClientAuthRemove,
    ReplyOnionClientAuthView,
    ReplyPostDescriptor,
    ReplyProtocolInfo,
    ReplyQuit,
    ReplyRedirectStream,
    ReplyResetConf,
    ReplyResolve,
    ReplySaveConf,
    ReplySetCircuitPurpose,
    ReplySetConf,
    ReplySetEvents,
    ReplySignal,
    ReplyTakeOwnership,
)
from .structures import (
    CircuitPurpose,
    DescriptorPurpose,
    HiddenServiceAddress,
    HiddenServiceAddressV3,
    HsDescClientAuth,
    LongServerName,
    OnionClientAuthFlags,
    OnionServiceFlags,
    OnionServiceKeyType,
    OnionServiceNewKeyStruct,
    ReplyDataAuthChallenge,
    Signal,
    StreamCloseReasonInt,
    VirtualPortTarget,
)
from .utils import Message, messages_from_stream

if TYPE_CHECKING:
    from collections.abc import (  # noqa: F401
        Mapping,
        MutableMapping,
        MutableSequence,
        Sequence,
        Set as AbstractSet,
    )
    from types import TracebackType
    from typing import Self

    from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
    from cryptography.hazmat.primitives.asymmetric.x25519 import (
        X25519PrivateKey,
        X25519PublicKey,
    )

    from .types import AnyHost


_EVENTS_TOR = frozenset(EventWord)
_EVENTS_INTERNAL = frozenset(EventWordInternal)
logger = logging.getLogger(__package__)

#: Alias for event callbacks registered with :meth:`Controller.add_event_handler`.
EventCallbackType: TypeAlias = Callable[[Event], Awaitable[None] | None]


class Controller:
    """Client controller for Tor's control socket."""

    def __init__(self, connector: ControlConnector) -> None:
        """
        Initialize a new controller from a provided :class:`.ControlConnector`.

        Notes:
            You may want to alternatively use one of the following methods:
                - :meth:`Controller.from_path`
                - :meth:`Controller.from_port`

        Args:
            connector: the connector to the control socket.

        """
        self._evt_callbacks = {}  # type: MutableMapping[str, list[EventCallbackType]]
        self._events_lock = asyncio.Lock()
        self._request_lock = asyncio.Lock()
        self._authenticated = False
        self._context = None  # type: AsyncExitStack | None
        self._connected = False
        self._connector = connector
        self._protoinfo = None  # type: ReplyProtocolInfo | None
        self._replies = None  # type: asyncio.Queue[Message | None] | None
        self._writer = None  # type: asyncio.StreamWriter | None

    async def __aenter__(self) -> Self:
        """
        Enter Controller's context, connect to the target.

        Raises:
            RuntimeError: when the context has already been entered.

        Returns:
            A connected controller (the same exact instance).

        """
        if self.entered:
            msg = 'Controller is already entered!'
            raise RuntimeError(msg)

        context = await AsyncExitStack().__aenter__()
        try:
            reader, writer = await self._connector.connect()
            context.push_async_callback(writer.wait_closed)
            context.callback(writer.close)

            replies = asyncio.Queue()  # type: asyncio.Queue[Message | None]
            rdtask = asyncio.create_task(
                self._reader_task(reader, replies),
                name='aiostem.controller.reader',
            )

            async def cancel_reader(task: asyncio.Task[None]) -> None:
                task.cancel('Controller is closing')
                await asyncio.gather(task, return_exceptions=True)

            context.push_async_callback(cancel_reader, rdtask)
        except BaseException:
            await context.aclose()
            raise
        else:
            self._context = context
            self._connected = True
            self._replies = replies
            self._writer = writer
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool:
        """
        Exit the controller's context and close the underlying socket.

        Returns:
            :obj:`False` to let any exception flow through the call stack.

        """
        context = self._context
        try:
            if context is not None:
                # Can arise while closing an underlying UNIX socket
                with suppress(BrokenPipeError):
                    await context.__aexit__(exc_type, exc_val, exc_tb)
        finally:
            self._authenticated = False
            self._connected = False
            self._context = None
            self._evt_callbacks.clear()
            self._protoinfo = None
            self._replies = None
            self._writer = None

        # Do not prevent the original exception from going further.
        return False

    @classmethod
    def from_port(
        cls,
        host: str = DEFAULT_CONTROL_HOST,
        port: int = DEFAULT_CONTROL_PORT,
    ) -> Controller:
        """
        Create a new controller for a remote TCP host/port.

        USE EXAMPLE::

            async with Controller.from_port('10.0.0.1', 9051) as controller:
                await controller.authenticate('password')
                ...

        Args:
            host: ip address or hostname to the control host.
            port: TCP port to connect to.

        Returns:
            A controller for the target TCP host and port.

        """
        return cls(ControlConnectorPort(host, port))

    @classmethod
    def from_path(cls, path: str = DEFAULT_CONTROL_PATH) -> Controller:
        """
        Create a new controller for a local unix socket.

        USE EXAMPLE::

            async with Controller.from_path('/run/tor/control.sock') as controller:
                await controller.authenticate()
                ...

        Args:
            path: path to the unix socket on the file system.

        Returns:
            A controller for the target unix socket.

        """
        return cls(ControlConnectorPath(path))

    @property
    def authenticated(self) -> bool:
        """Tell whether we are correctly authenticated."""
        return bool(self.connected and self._authenticated)

    @property
    def connected(self) -> bool:
        """Tell whether we are connected to the remote socket."""
        return bool(self._connected and self._writer is not None and self._replies is not None)

    @property
    def entered(self) -> bool:
        """Tell whether the context manager is entered."""
        return bool(self._context is not None)

    @staticmethod
    def _str_event_to_enum(event: str) -> EventWord | EventWordInternal:
        """
        Convert a textual event name to its enum equivalent.

        Args:
            event: a textual representation of the event.

        Raises:
            CommandError: when the event name does not exit.

        Returns:
            The corresponding enum event.

        """
        event = event.upper()
        if event in _EVENTS_TOR:
            return EventWord(event)
        if event in _EVENTS_INTERNAL:
            return EventWordInternal(event)
        msg = f"Unknown event '{event}'"
        raise CommandError(msg)

    async def _notify_disconnect(self) -> None:
        """
        Generate a `DISCONNECT` event.

        The goal here is to tell everyone that we are now disconnected from the remote socket.
        This is a fake event simply used to call all the registered callbacks and provide a
        way to gently tell the end user that we are no longer capable of handling anything.
        """
        await self._on_event_received(Message(status=650, header='DISCONNECT'))

    async def _on_event_received(self, message: Message) -> None:
        """
        Handle a newly received event.

        This method finds and call the appropriate callbacks, if any.

        Args:
            message: the raw event message.

        """
        try:
            event = event_from_message(message)
        except Exception:
            logger.exception('Unable to handle a received event.')
        else:
            keyword = message.keyword
            for callback in self._evt_callbacks.get(keyword, []):
                # We do not care about exceptions in the event callback.
                try:
                    coro = callback(event)
                    if asyncio.iscoroutine(coro):
                        await coro
                except Exception:
                    logger.exception("Error while handling callback for '%s'", keyword)

    async def _reader_task(
        self,
        reader: asyncio.StreamReader,
        replies: asyncio.Queue[Message | None],
    ) -> None:
        """
        Read messages from the control socket and dispatch them.

        Args:
            reader: raw StreamReader from :mod:`asyncio`.
            replies: the queue of command replies.

        """
        try:
            async for message in messages_from_stream(reader):
                if message.is_event:
                    await self._on_event_received(message)
                else:
                    await replies.put(message)
        finally:
            self._connected = False
            # This is needed here because we may be stuck waiting on a reply.
            with suppress(asyncio.QueueFull):
                replies.put_nowait(None)
            await self._notify_disconnect()

    async def add_event_handler(
        self,
        event: EventWord | EventWordInternal | str,
        callback: EventCallbackType,
    ) -> None:
        """
        Register a callback function to be called when an event message is received.

        Notes:
            - A special event :attr:`~.event.EventWordInternal.DISCONNECT` is handled
              internally by this library and can be registered here to be notified of any
              disconnection from the control socket.
            - Multiple callbacks can be set for a single event.
              If so, they are called in the order they were registered.

        USE EXAMPLE::

            def client_status_callback(event: EventStatusClient):
                print(event)

            async with Controller.from_path('/run/tor/control.sock') as controller:
                await controller.authenticate()
                await controller.add_event_handler('STATUS_CLIENT', client_status_callback)
                ...

        See Also:
            https://spec.torproject.org/control-spec/replies.html#asynchronous-events

        Args:
            event: Name of the event linked to the callback.
            callback: A function or coroutine to be called when the event occurs.

        Raises:
            CommandError: When the event name does not exit.
            ReplyStatusError: When the event could not be registered.

        """
        if not isinstance(event, EventWord | EventWordInternal):
            event = self._str_event_to_enum(event)

        async with self._events_lock:
            listeners = self._evt_callbacks.setdefault(event.value, [])
            try:
                # Tell Tor that we are now interested in this event.
                if not len(listeners) and isinstance(event, EventWord):
                    keys = frozenset(self._evt_callbacks.keys())
                    reals = keys.difference(EventWordInternal)
                    events = frozenset(map(EventWord, reals))

                    reply = await self.set_events(events)
                    reply.raise_for_status()
            except BaseException:
                if not len(listeners):  # pragma: no branch
                    self._evt_callbacks.pop(event)
                raise
            else:
                listeners.append(callback)

    async def del_event_handler(
        self,
        event: EventWord | EventWordInternal | str,
        callback: EventCallbackType,
    ) -> None:
        """
        Unregister a previously registered callback function.

        Args:
            event: Name of the event linked to the callback.
            callback: A function or coroutine to be removed from the event list.

        """
        if not isinstance(event, EventWord | EventWordInternal):
            event = self._str_event_to_enum(event)

        async with self._events_lock:
            listeners = self._evt_callbacks.get(event, [])
            if callback in listeners:
                backup_listeners = listeners.copy()
                listeners.remove(callback)
                try:
                    if not len(listeners):
                        self._evt_callbacks.pop(event.value)
                        if isinstance(event, EventWord):
                            keys = frozenset(self._evt_callbacks.keys())
                            reals = keys.difference(EventWordInternal)
                            events = frozenset(map(EventWord, reals))
                            await self.set_events(events)
                except BaseException:
                    # Restore the original callbacks on error.
                    self._evt_callbacks[event.value] = backup_listeners
                    raise

    async def request(self, command: Command) -> Message:
        """
        Send any kind of command to the controller.

        This method is the underlying call of any other command.

        It can be used to send custom subclass of :class:`.Command`, and get the
        raw :class:`.Message` corresponding to the response. This :class:`.Message`
        can then be parsed by an appropriate :class:`.Reply`.

        Important:
            A single command can run at any time due to an internal lock.

        Args:
            command: The command we want to send to Tor.

        Raises:
            ControllerError: When the controller is not connected.

        Returns:
            The corresponding reply message from the remote daemon.

        """
        async with self._request_lock:
            # if self._replies is None or self._writer is None:
            if not self.connected:
                msg = 'Controller is not connected!'
                raise ControllerError(msg)

            # Casts are valid here since we check `self.connected`.
            replies = cast('asyncio.Queue[Message | None]', self._replies)
            writer = cast('asyncio.StreamWriter', self._writer)

            frame = command.serialize()
            writer.write(frame.encode('ascii'))
            await writer.drain()

        resp = await replies.get()
        replies.task_done()

        # This is very hard to test here, let's not cover these few lines.
        if resp is None:  # pragma: no cover
            msg = 'Controller has disconnected!'
            raise ControllerError(msg)
        return resp

    async def add_onion(
        self,
        key: Ed25519PrivateKey | RSAPrivateKey | OnionServiceNewKeyStruct | str,
        ports: Sequence[VirtualPortTarget | str],
        *,
        client_auth: Sequence[HsDescClientAuth] = [],
        client_auth_v3: Sequence[X25519PublicKey] = [],
        flags: AbstractSet[OnionServiceFlags] = frozenset(),
        generate_locally: bool = True,
        max_streams: int | None = None,
    ) -> ReplyAddOnion:
        """
        Create a new onion service.

        This can either be created from an existing private key or ask Tor generate
        a new one. You must provide at least one ``port`` for this command to succeed.

        Args:
            key: The onion service private key or type of private key to generate.
            ports: List of virtual ports to assign to this onion service.

        Keyword Args:
            client_auth: List of authorized clients (for V2 onions).
            client_auth_v3: List of authorized clients (for V3 onions).
            flags: Set of additional flags attached to this service.
            generate_locally: Generate onion keys locally instead (default).
            max_streams: Optional max number of streams attached to the rendezvous circuit.

        Returns:
            The reply from the Tor server.

        """
        adapter = CommandAddOnion.adapter()
        command = adapter.validate_python(
            {
                'key': key,
                'ports': ports,
                'client_auth': client_auth,
                'client_auth_v3': client_auth_v3,
                'flags': flags,
                'max_streams': max_streams,
            },
        )
        # Make sure the V3AUTH flag is set when necessary.
        if len(command.client_auth_v3):
            command.flags.add(OnionServiceFlags.V3AUTH)

        generated = None  # type: Ed25519PrivateKey | RSAPrivateKey | None
        if generate_locally and isinstance(command.key, OnionServiceNewKeyStruct):
            match command.key.key_type:
                case OnionServiceKeyType.RSA1024:
                    # These are/were the common parameters for Onion V2.
                    generated = rsa.generate_private_key(65537, key_size=1024)  # noqa: S505
                    command.key = generated
                case _:
                    generated = Ed25519PrivateKey.generate()
                    command.key = generated

        message = await self.request(command)
        reply = ReplyAddOnion.from_message(message)

        # Simply copy the key generated locally.
        if (
            generated is not None
            and reply.data is not None
            and OnionServiceFlags.DISCARD_PK not in command.flags
        ):
            reply.data.key = generated
        return reply

    async def attach_stream(
        self,
        stream: int,
        circuit: int,
        *,
        hop: int | None = None,
    ) -> ReplyAttachStream:
        """
        Inform the server that the specified stream should be associated with a circuit.

        Args:
            stream: The stream to associate to the provided circuit.
            circuit: The circuit identifier to attach the stream onto.

        Keyword Args:
            hop: When specified, Tor choose the HopNumth hop in the circuit as the exit node.

        Returns:
            A simple reply where only the status is relevant.

        """
        command = CommandAttachStream(
            stream=stream,
            circuit=circuit,
            hop=hop,
        )
        message = await self.request(command)
        return ReplyAttachStream.from_message(message)

    async def auth_challenge(self, nonce: bytes | str | None = None) -> ReplyAuthChallenge:
        """
        Start the authentication for :attr:`~.structures.AuthMethod.SAFECOOKIE`.

        When no ``nonce`` is provided, once is generated and provided back in the reply.
        While this is obviously not part of the original reply from the server, it is
        added to the reply structure for convenience.

        Warning:
            This method is not meant to be called by the end-user but is rather
            used internally by :meth:`authenticate`.

        Note:
            This command can be sent while not authenticated (but only once).

        Args:
            nonce: 32 random bytes (optional).

        Returns:
            An authentication challenge reply.

        """
        if nonce is None:
            nonce = CommandAuthChallenge.generate_nonce()

        command = CommandAuthChallenge(nonce=nonce)
        message = await self.request(command)
        reply = ReplyAuthChallenge.from_message(message)
        if reply.data is not None:
            reply.data.client_nonce = command.nonce
        return reply

    async def authenticate(self, password: str | None = None) -> ReplyAuthenticate:
        """
        Authenticate to Tor's controller.

        Note:
            Available authentications are provided by :meth:`protocol_info`.

        Important:
            Authentication methods are tried in the following order (when available):
                - ``NULL``: authentication is automatically granted
                - ``HASHEDPASSWORD``: password authentication (when a password is provided)
                - ``SAFECOOKIE``: proof that we can read the cookie file
                - ``COOKIE``: provide the content of the cookie file

        See Also:
            :class:`.structures.AuthMethod`

        Args:
            password: Optional password for method
                :attr:`~.structures.AuthMethod.HASHEDPASSWORD`.

        Raises:
            ControllerError: When no known authentication method was found.

        Returns:
            The authentication reply (you should check the status here).

        """
        protoinfo = await self.protocol_info()
        protoinfo.raise_for_status()

        methods = set()  # type: set[str]
        if protoinfo.data is not None:  # pragma: no branch
            methods.update(protoinfo.data.auth_methods)

        # No password was provided, we can't authenticate with this method.
        if password is None:
            methods.discard('HASHEDPASSWORD')

        # Here we suppose that the user prefers a password authentication when
        # a password is provided (the cookie file may not be readable).
        if 'NULL' in methods:
            token = None  # type: bytes | None
        elif 'HASHEDPASSWORD' in methods:
            token = password.encode()  # type: ignore[union-attr]
        elif 'SAFECOOKIE' in methods:
            cookie = await protoinfo.read_cookie_file()
            reply = await self.auth_challenge()
            reply.raise_for_status()
            # We know that reply.data is set because `raise_for_status` would have raised.
            challenge = cast('ReplyDataAuthChallenge', reply.data)
            challenge.raise_for_server_hash_error(cookie)
            token = challenge.build_client_hash(cookie)
        elif 'COOKIE' in methods:
            token = await protoinfo.read_cookie_file()
        else:
            msg = 'No compatible authentication method found!'
            raise ControllerError(msg)

        command = CommandAuthenticate(token=token)
        message = await self.request(command)
        self._authenticated = message.is_success
        return ReplyAuthenticate.from_message(message)

    async def close_circuit(
        self,
        circuit: int,
        *,
        if_unused: bool = False,
    ) -> ReplyCloseCircuit:
        """
        Tell the server to close the specified circuit.

        Args:
            circuit: The circuit identifier to close.

        Keyword Args:
            if_unused: Close the circuit only if it is unused.

        Returns:
            A simple reply where only the status is relevant.

        """
        command = CommandCloseCircuit(circuit=circuit, if_unused=if_unused)
        message = await self.request(command)
        return ReplyCloseCircuit.from_message(message)

    async def close_stream(
        self,
        stream: int,
        reason: StreamCloseReasonInt,
    ) -> ReplyCloseStream:
        """
        Tell the server to close the specified stream.

        Args:
            stream: The stream identifier to close.
            reason: The provided reason why this stream should be closed.

        Returns:
            A simple reply where only the status is relevant.

        """
        command = CommandCloseStream(stream=stream, reason=reason)
        message = await self.request(command)
        return ReplyCloseStream.from_message(message)

    async def del_onion(self, address: HiddenServiceAddress | str) -> ReplyDelOnion:
        """
        Request the deletion of a hidden service we control.

        Args:
            address: The hidden service address to delete.

        Returnse
            A simple del_onion reply where only the status is relevant.

        """
        adapter = CommandDelOnion.adapter()
        command = adapter.validate_python({'address': address})
        message = await self.request(command)
        return ReplyDelOnion.from_message(message)

    async def drop_guards(self) -> ReplyDropGuards:
        """
        Tell the server to drop all guard nodes.

        Warning:
            Do not invoke this command lightly; it can increase vulnerability
            to tracking attacks over time.

        Returns:
            A simple drop-guards reply where only the status is relevant.

        """
        command = CommandDropGuards()
        message = await self.request(command)
        return ReplyDropGuards.from_message(message)

    async def drop_ownership(self) -> ReplyDropOwnership:
        """
        Relinquish ownership of this control connection.

        See Also:
            https://spec.torproject.org/control-spec/commands.html#dropownership

        Hint:
            This ownership can be taken using :meth:`take_ownership`.

        Returns:
            A simple drop-ownership reply where only the status is relevant.

        """
        command = CommandDropOwnership()
        message = await self.request(command)
        return ReplyDropOwnership.from_message(message)

    async def drop_timeouts(self) -> ReplyDropTimeouts:
        """
        Tells the server to drop all circuit build times.

        Warning:
            Do not invoke this command lightly; it can increase vulnerability
            to tracking attacks over time.

        Note:
            Tor also emits the ``BUILDTIMEOUT_SET RESET`` event right after the reply.

        Returns:
            A simple drop-timeouts reply where only the status is relevant.

        """
        command = CommandDropTimeouts()
        message = await self.request(command)
        return ReplyDropTimeouts.from_message(message)

    async def extend_circuit(
        self,
        circuit: int,
        servers: Sequence[LongServerName | str],
        *,
        purpose: CircuitPurpose | str | None = None,
    ) -> ReplyExtendCircuit:
        """
        Extend the provided circuit to the provided servers.

        Args:
            circuit: The circuit identifier to extend to (``0`` for a new circuit).
            servers: List of servers to extend the circuit to.

        Keyword Args:
            purpose: Optional circuit purpose.

        Returns:
            A circuit reply containing the circuit number.

        """
        adapter = CommandExtendCircuit.adapter()
        command = adapter.validate_python(
            {
                'circuit': circuit,
                'servers': servers,
                'purpose': purpose,
            },
        )
        message = await self.request(command)
        return ReplyExtendCircuit.from_message(message)

    async def get_conf(self, *args: str) -> ReplyGetConf:
        """
        Request the value of zero or move configuration variable(s).

        Note that you can request the same key multiple times, and some configuration
        entries can provide multiple values. When any of this happens, the result
        dictionary provides a :class:`~typing.Sequence` of strings as its value.

        See Also:
            https://spec.torproject.org/control-spec/commands.html#getconf

        Args:
            args: A list of configuration variables to request.

        Returns:
            A reply containing the corresponding values (when successful).

        """
        command = CommandGetConf(keywords=[*args])
        message = await self.request(command)
        return ReplyGetConf.from_message(message)

    async def get_info(self, *args: str) -> ReplyGetInfo:
        """
        Request for Tor daemon information.

        Note that you can request the same key multiple times. When this happens, the result
        dictionary provides a :class:`~typing.Sequence` of strings as its value.

        See Also:
            https://spec.torproject.org/control-spec/commands.html#getinfo

        Args:
            args: A list of information data to request.

        Returns:
            A reply containing the corresponding values (when successful).

        """
        command = CommandGetInfo(keywords=[*args])
        message = await self.request(command)
        return ReplyGetInfo.from_message(message)

    async def hs_fetch(
        self,
        address: HiddenServiceAddress | str,
        servers: Sequence[LongServerName | str] = [],
    ) -> ReplyHsFetch:
        """
        Request a hidden service descriptor fetch.

        The result does not contain the descriptor, which is provided asynchronously through
        events such as :attr:`~.EventWord.HS_DESC` or :attr:`~.EventWord.HS_DESC_CONTENT`.

        Args:
            address: The hidden service address to request.
            servers: An optional list of servers to query.

        Returns:
            A simple hsfetch reply where only the status is relevant.

        """
        adapter = CommandHsFetch.adapter()
        command = adapter.validate_python(
            {
                'address': address,
                'servers': servers,
            }
        )
        message = await self.request(command)
        return ReplyHsFetch.from_message(message)

    async def hs_post(
        self,
        descriptor: str,
        *,
        address: HiddenServiceAddressV3 | str | None = None,
        servers: Sequence[LongServerName | str] = [],
    ) -> ReplyHsPost:
        """
        Ask Tor to upload the provided descriptor.

        Important:
            ``address`` is mandatory when uploading a v3 descriptor.

        Warning:
            When an invalid descriptor is provided and no address is provided,
            tor does not provide an answer to this command, make sure to wrap
            this function call with a timeout!

        Args:
            descriptor: The descriptor content to upload.

        Keyword Args:
            address: For v3 only, the hidden service address.
            servers: List of servers to upload the descriptor to.

        Returns:
            A simple reply where only the status is relevant.

        """
        adapter = CommandHsPost.adapter()
        command = adapter.validate_python(
            {
                'descriptor': descriptor,
                'servers': servers,
                'address': address,
            },
        )
        message = await self.request(command)
        return ReplyHsPost.from_message(message)

    async def load_conf(self, text: str) -> ReplyLoadConf:
        """
        Upload and replace the content of a config file.

        This command allows a controller to upload the text of a config file to Tor over
        the control port. This config file is then loaded as if it had been read from disk.

        See Also:
            https://spec.torproject.org/control-spec/commands.html#loadconf

        Returns:
            A simple reply with only a status.

        """
        command = CommandLoadConf(text=text)
        message = await self.request(command)
        return ReplyLoadConf.from_message(message)

    async def map_address(self, addresses: Mapping[AnyHost, AnyHost]) -> ReplyMapAddress:
        """
        Map provided addresses with their replacement.

        The client tells the server that future SOCKS requests for connections to any
        original address provided here should be replaced with a connection to the
        specified replacement address.

        The client may decline to provide a replacement address and instead provide
        a special address. This means that the server should choose the original
        address itself.

        - For IPv4: ``0.0.0.0``
        - For IPv6: ``::0``
        - For hostname: ``.``

        Mapping values can be read using :meth:`get_info` with ``address-mappings/control``.

        See Also:
            https://spec.torproject.org/control-spec/commands.html#mapaddress

        Args:
            addresses: A map of addresses to remap on socks requests.

        Returns:
            A list of individual replies for each map request.

            Note that some values can be rejected and others can be accepted, which means
            that you should check each individual value.

        """
        command = CommandMapAddress()
        command.addresses.update(addresses)
        message = await self.request(command)
        return ReplyMapAddress.from_message(message)

    async def onion_client_auth_add(
        self,
        address: HiddenServiceAddressV3 | str,
        key: X25519PrivateKey,
        *,
        flags: AbstractSet[OnionClientAuthFlags] = frozenset(),
        nickname: str | None = None,
    ) -> ReplyOnionClientAuthAdd:
        """
        Add a new client authorization for an existing hidden service.

        Args:
            address: Onion service address to add a client authorization for.
            key: The private X25519 private key used to access this service.

        Keyword Args:
            flags: Set of additional flags.
            nickname: Client name associated with the provided key.

        Returns:
            A simple reply where only the status is relevant.

        """
        adapter = CommandOnionClientAuthAdd.adapter()
        command = adapter.validate_python(
            {
                'address': address,
                'key': key,
                'flags': flags,
                'nickname': nickname,
            }
        )
        message = await self.request(command)
        return ReplyOnionClientAuthAdd.from_message(message)

    async def onion_client_auth_remove(
        self,
        address: HiddenServiceAddressV3 | str,
    ) -> ReplyOnionClientAuthRemove:
        """
        Remove a client authorization for an existing hidden service.

        Args:
            address: Onion service address to remove a client authorization from.

        Returns:
            A simple reply where only the status is relevant.

        """
        adapter = CommandOnionClientAuthRemove.adapter()
        command = adapter.validate_python({'address': address})
        message = await self.request(command)
        return ReplyOnionClientAuthRemove.from_message(message)

    async def onion_client_auth_view(
        self,
        address: HiddenServiceAddressV3 | str | None = None,
    ) -> ReplyOnionClientAuthView:
        """
        List authorization clients for the provided hidden service or all services.

        Args:
            address: Optional onion service address to list authorization for.

        Returns:
            A simple reply where only the status is relevant.

        """
        adapter = CommandOnionClientAuthView.adapter()
        command = adapter.validate_python({'address': address})
        message = await self.request(command)
        return ReplyOnionClientAuthView.from_message(message)

    async def post_descriptor(
        self,
        descriptor: str,
        *,
        cache: bool | None = None,
        purpose: DescriptorPurpose | None = None,
    ) -> ReplyPostDescriptor:
        """
        Inform the server about a new router descriptor.

        Args:
            descriptor: The router descriptor content.

        Keyword Args:
            cache: Whether to cache the provided descriptor internally.
            purpose: The purpose of the descriptor (default is ``GENERAL``).

        Returns:
            A simple reply where only the status is relevant.

        """
        adapter = CommandPostDescriptor.adapter()
        command = adapter.validate_python(
            {
                'descriptor': descriptor,
                'purpose': purpose,
                'cache': cache,
            },
        )
        message = await self.request(command)
        return ReplyPostDescriptor.from_message(message)

    async def protocol_info(self, version: int | None = None) -> ReplyProtocolInfo:
        """
        Get control protocol information from Tor.

        This command is performed as part of the authentication process in order to find out
        all supported authentication methods (see :class:`~.structures.AuthMethod`).

        The ``version`` is supposed to set to ``1`` but Tor currently does not care.

        Note:
            The command result is cached when unauthenticated as we can only send this
            command once in this situation.

        See Also:
            https://spec.torproject.org/control-spec/commands.html#protocolinfo

        Args:
            version: Protocol version to ask for when provided.

        Returns:
            A completed protocol info reply from Tor.

        """
        if self.authenticated or self._protoinfo is None:
            command = CommandProtocolInfo(version=version)
            message = await self.request(command)
            self._protoinfo = ReplyProtocolInfo.from_message(message)
        return self._protoinfo

    async def quit(self) -> ReplyQuit:
        """
        Tells the server to hang up on this controller connection.

        Returns:
            A simple quit reply where only the status is relevant.

        """
        message = await self.request(CommandQuit())
        return ReplyQuit.from_message(message)

    async def redirect_stream(
        self,
        stream: int,
        address: AnyHost,
        *,
        port: int | None = None,
    ) -> ReplyRedirectStream:
        """
        Tell the server to change the exit address on the specified stream.

        Args:
            stream: The stream identifier to redirect.
            address: Destination address to redirect it to.

        Keyword Args:
            port: Optional port to redirect the stream to.

        Returns:
            A simple reply where only the status is relevant.

        """
        adapter = CommandRedirectStream.adapter()
        command = adapter.validate_python(
            {
                'stream': stream,
                'address': address,
                'port': port,
            },
        )
        message = await self.request(command)
        return ReplyRedirectStream.from_message(message)

    async def reset_conf(
        self,
        items: Mapping[str, MutableSequence[int | str] | int | str | None],
    ) -> ReplyResetConf:
        """
        Change or reset configuration entries on the remote server.

        Notes:
            - When :obj:`None` is provided, all values are reset to their default.
            - When :class:`list`, multiple values are assigned.

        See Also:
            https://spec.torproject.org/control-spec/commands.html#resetconf

        Args:
            items: a map of configuration entries to apply or reset.

        Returns:
            A simple resetconf reply where only the status is relevant.

        """
        command = CommandResetConf()
        command.values.update(items)
        message = await self.request(command)
        return ReplyResetConf.from_message(message)

    async def resolve(
        self,
        addresses: Sequence[AnyHost],
        *,
        reverse: bool = False,
    ) -> ReplyResolve:
        """
        Launch a remote hostname lookup request for every specified request.

        Note:
            The result is not provided along with the reply here but can be caught
            using the :attr:`~.EventWord.ADDRMAP`.

        See Also:
            https://spec.torproject.org/control-spec/commands.html#resolve

        Args:
            addresses: List of addresses to launch a resolve request for.

        Keyword Arguments:
            reverse: Whether to perform a reverse DNS lookup.

        Returns:
            A simple resolve reply where only the status is relevant.

        """
        command = CommandResolve(reverse=reverse)
        command.addresses.extend(addresses)
        message = await self.request(command)
        return ReplyResolve.from_message(message)

    async def save_conf(self, *, force: bool = False) -> ReplySaveConf:
        """
        Instructs the server to write out its configuration options into ``torrc``.

        If ``%include`` is used on ``torrc``, ``SAVECONF`` will not write the configuration
        to disk.  When set, the configuration will be overwritten even if %include is used.
        You can find out whether this flag is needed using ``config-can-saveconf`` on
        :class:`.CommandGetInfo`.

        Keyword Arguments:
            force: force write the configuration to disk.

        See Also:
            https://spec.torproject.org/control-spec/commands.html#saveconf

        Returns:
            A simple reply with only a status.

        """
        command = CommandSaveConf(force=force)
        message = await self.request(command)
        return ReplySaveConf.from_message(message)

    async def set_circuit_purpose(
        self,
        circuit: int,
        purpose: CircuitPurpose | str,
    ) -> ReplySetCircuitPurpose:
        """
        Change the circuit purpose.

        Args:
            circuit: The circuit identifier to change the purpose.
            purpose: New purpose for the provided circuit.

        Returns:
            A simple reply with only a status.

        """
        adapter = CommandSetCircuitPurpose.adapter()
        command = adapter.validate_python(
            {
                'circuit': circuit,
                'purpose': purpose,
            },
        )
        message = await self.request(command)
        return ReplySetCircuitPurpose.from_message(message)

    async def set_conf(
        self,
        items: Mapping[str, MutableSequence[int | str] | int | str | None],
    ) -> ReplySetConf:
        """
        Change configuration entries on the remote server.

        Notes:
            - When :obj:`None` is provided, all values are removed.
            - When :class:`list`, multiple values are assigned.

        See Also:
            https://spec.torproject.org/control-spec/commands.html#setconf

        Args:
            items: a map of new configuration entries to apply or clear.

        Returns:
            A simple setconf reply where only the status is relevant.

        """
        command = CommandSetConf()
        command.values.update(items)
        message = await self.request(command)
        return ReplySetConf.from_message(message)

    async def set_events(self, events: AbstractSet[EventWord]) -> ReplySetEvents:
        """
        Set the list of events that we subscribe to.

        Warning:
            This method should not probably be called by the end-user.
            Please see :meth:`add_event_handler` instead.

        Args:
            events: a set of events to subscribe to.

        Returns:
            A simple setevents reply where only the status is relevant.

        """
        command = CommandSetEvents()
        command.events.update(events)
        message = await self.request(command)
        return ReplySetEvents.from_message(message)

    async def signal(self, signal: Signal | str) -> ReplySignal:
        """
        Send a signal to the controller.

        See Also:
            https://spec.torproject.org/control-spec/commands.html#signal

        Args:
            signal: name of the signal to send.

        Returns:
            A simple signal reply where only the status is relevant.

        """
        adapter = CommandSignal.adapter()
        command = adapter.validate_python({'signal': signal})
        message = await self.request(command)
        return ReplySignal.from_message(message)

    async def take_ownership(self) -> ReplyTakeOwnership:
        """
        Instructs Tor to shut down when this control connection is closed.

        Hint:
            This ownership can be dropped with :meth:`drop_ownership`.

        See Also:
            https://spec.torproject.org/control-spec/commands.html#takeownership

        Returns:
            A simple take-ownership reply where only the status is relevant.

        """
        command = CommandTakeOwnership()
        message = await self.request(command)
        return ReplyTakeOwnership.from_message(message)



from __future__ import annotations

import secrets
from base64 import b32encode, b64encode
from collections.abc import Sequence
from datetime import datetime
from ipaddress import IPv4Address, IPv6Address
from typing import Annotated, ClassVar

import pytest
from cryptography.hazmat.primitives.asymmetric.ed25519 import (
    Ed25519PrivateKey,
    Ed25519PublicKey,
)
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PrivateKey
from pydantic import BaseModel, TypeAdapter, ValidationError

from aiostem.exceptions import CryptographyError, ReplySyntaxError
from aiostem.structures import (
    Ed25519CertExtension,
    Ed25519CertExtensionSigningKey,
    Ed25519CertExtensionUnkown,
    Ed25519CertificateV1,
    HiddenServiceAddress,
    HiddenServiceAddressV2,
    HiddenServiceAddressV3,
    HsDescAuthCookie,
    HsDescAuthTypeInt,
    HsDescV3Layer2,
    LinkSpecifierStruct,
    LogSeverity,
    LongServerName,
    OnionClientAuthKey,
    OnionServiceKey,
    OnionServiceKeyStruct,
    StreamTarget,
    TcpAddressPort,
    TrLinkSpecifierList,
    _parse_block,
)
from aiostem.utils import TrEd25519PrivateKey

HiddenServiceAdapter = TypeAdapter(HiddenServiceAddress)
HiddenServiceAdapterV2 = TypeAdapter(HiddenServiceAddressV2)
HiddenServiceAdapterV3 = TypeAdapter(HiddenServiceAddressV3)


class TestHiddenServiceV2:
    @pytest.mark.parametrize(
        'address',
        [
            HiddenServiceAdapterV2.validate_python('facebookcorewwwi.onion'),
            'facebookcorewwwi.onion',
            'facebookcorewwwi',
        ],
    )
    def test_valid_domains(self, address):
        class Model(BaseModel):
            v: HiddenServiceAddressV2

        model = Model(v=address)
        assert model.v == 'facebookcorewwwi'

    @pytest.mark.parametrize(
        ('address', 'errtype'),
        [
            ('facebookcorewww1', 'string_pattern_mismatch'),
            ('facebookcorewww.onion', 'string_pattern_mismatch'),
            ('facebookcorewww', 'string_too_short'),
        ],
    )
    def test_invalid_domains(self, address, errtype):
        class Model(BaseModel):
            v: HiddenServiceAddressV2

        with pytest.raises(ValidationError, match=f'type={errtype}') as exc:
            Model(v=address)

        # Here we have two errors since we are neither and instance nor compatible.
        assert len(exc.value.errors()) == 2, exc.value.errors()
        error = exc.value.errors()[0]
        assert error['type'] == 'is_instance_of', address
        error = exc.value.errors()[1]
        assert error['type'] == errtype, address


class TestHiddenServiceV3:
    @pytest.mark.parametrize(
        'address',
        [
            HiddenServiceAdapterV3.validate_python(
                'facebookcooa4ldbat4g7iacswl3p2zrf5nuylvnhxn6kqolvojixwid.onion',
            ),
            'facebookcooa4ldbat4g7iacswl3p2zrf5nuylvnhxn6kqolvojixwid.onion',
            'facebookcooa4ldbat4g7iacswl3p2zrf5nuylvnhxn6kqolvojixwid',
        ],
    )
    def test_valid_domains(self, address):
        class Model(BaseModel):
            v: HiddenServiceAddressV3

        model = Model(v=address)
        assert model.v == 'facebookcooa4ldbat4g7iacswl3p2zrf5nuylvnhxn6kqolvojixwid'

    def test_ed25519_public_key(self):
        """Check that the public ed25519 key is correct."""
        address = 'facebookcooa4ldbat4g7iacswl3p2zrf5nuylvnhxn6kqolvojixwid'
        adapter = TypeAdapter(HiddenServiceAddressV3)
        onion = adapter.validate_python(address)
        assert isinstance(onion, HiddenServiceAddressV3)

        pubkey = onion.public_key
        assert isinstance(pubkey, Ed25519PublicKey)

        raw = pubkey.public_bytes_raw()
        prefix = b32encode(raw).decode('ascii').lower()[:31]
        assert address.startswith(prefix) is True

    @pytest.mark.parametrize(
        ('address', 'errtype'),
        [
            (
                'facebookcooa4ldbat4g7iacswl3p2zrf5nuylvnhxn6kqolvojixw1d',
                'string_pattern_mismatch',
            ),
            (
                'facebookcooa4ldbat4g7iacswl3p2zrf5nuylvnhxn6kqol',
                'string_too_short',
            ),
            (
                'facebookcooa4ldbat4g7iacswl3p2zrf5nuylvnhxn6kqolvojixwib',
                'invalid_onion_v3',
            ),
            (
                'facebookcooa4ldbat4g7iacswl3p2zrf5nuylvnhxn6kqolvojixsad',
                'invalid_onion_v3',
            ),
        ],
    )
    def test_invalid_domains(self, address, errtype):
        class Model(BaseModel):
            v: HiddenServiceAddressV3

        with pytest.raises(ValidationError, match=f'type={errtype}') as exc:
            Model(v=address)

        assert len(exc.value.errors()) == 2, exc.value.errors()
        error = exc.value.errors()[0]
        assert error['type'] == 'is_instance_of', address
        error = exc.value.errors()[1]
        assert error['type'] == errtype, address


class TestHiddenService:
    @pytest.mark.parametrize(
        ('address', 'version'),
        [
            (HiddenServiceAddressV2.from_string('facebookcorewwwi.onion'), 2),
            ('facebookcorewwwi.onion', 2),
            ('facebookcorewwwi', 2),
            (
                HiddenServiceAddressV3.from_string(
                    'facebookcooa4ldbat4g7iacswl3p2zrf5nuylvnhxn6kqolvojixwid.onion'
                ),
                3,
            ),
            ('facebookcooa4ldbat4g7iacswl3p2zrf5nuylvnhxn6kqolvojixwid.onion', 3),
            ('facebookcooa4ldbat4g7iacswl3p2zrf5nuylvnhxn6kqolvojixwid', 3),
        ],
    )
    def test_valid_domains(self, address, version):
        hsaddr = HiddenServiceAdapter.validate_python(address)
        assert version == hsaddr.VERSION

    @pytest.mark.parametrize(
        'address',
        [
            'facebookcorewww.onion',
            'facebookcorewww',
            'facebookcooa4ldbat4g7iacswl3p2zrf5nuylvnhxn6kqolvojixid.onion',
            'facebookcooa4ldbat4g7iacswl3p2zrf5nuylvnhxn6kqolvojixid',
            1234,
        ],
    )
    def test_invalid_domains(self, address):
        with pytest.raises(ValidationError, match='validation error'):
            HiddenServiceAdapter.validate_python(address)


HsDescAuthCookieAdapter = TypeAdapter(HsDescAuthCookie)


class TestHsDescAuthCookie:
    @pytest.mark.parametrize(
        ('raw', 'encoded', 'auth_type'),
        [
            ('GmYIu0EKkd5H6blpIFg3jQ', 'GmYIu0EKkd5H6blpIFg3jQA=', 1),
            ('GmYIu0EKkd5H6blpIFg3jQA=', 'GmYIu0EKkd5H6blpIFg3jQA=', 1),
            (
                bytes.fromhex('1a6608bb410a91de47e9b9692058378d00'),
                'GmYIu0EKkd5H6blpIFg3jQA=',
                1,
            ),
            (
                HsDescAuthCookie(
                    auth_type=HsDescAuthTypeInt.BASIC_AUTH,
                    cookie=bytes.fromhex('1a6608bb410a91de47e9b9692058378d'),
                ),
                'GmYIu0EKkd5H6blpIFg3jQA=',
                1,
            ),
            ('GmYIu0EKkd5H6blpIFg3jR', 'GmYIu0EKkd5H6blpIFg3jRA=', 2),
            ('GmYIu0EKkd5H6blpIFg3jRA=', 'GmYIu0EKkd5H6blpIFg3jRA=', 2),
        ],
    )
    def test_parse_then_encode(self, raw, encoded, auth_type):
        auth = HsDescAuthCookieAdapter.validate_python(raw)
        assert int(auth.auth_type) == auth_type
        serial = HsDescAuthCookieAdapter.dump_python(auth)
        assert serial == encoded

    @pytest.mark.parametrize(
        'auth_type',
        [
            HsDescAuthTypeInt.BASIC_AUTH,
            HsDescAuthTypeInt.STEALTH_AUTH,
        ],
    )
    def test_generate(self, auth_type):
        auth = HsDescAuthCookie.generate(auth_type)
        assert auth.auth_type == auth_type
        assert len(auth.cookie) == 16


class TestLogSeverity:
    ADAPTER = TypeAdapter(LogSeverity)
    TEST_CASES: ClassVar[Sequence[str]] = [
        LogSeverity.ERROR,
        'ERR',
        'ERROR',
        'Error',
        'err',
    ]

    @pytest.mark.parametrize('entry', TEST_CASES)
    def test_log_severity_with_values(self, entry):
        value = self.ADAPTER.validate_python(entry)
        assert value == LogSeverity.ERROR


LongServerNameAdapter = TypeAdapter(LongServerName)


class TestLongServerName:
    ADAPTER = TypeAdapter(LongServerName)

    @pytest.mark.parametrize(
        ('entry', 'nickname'),
        [
            (
                LongServerName(
                    fingerprint=bytes.fromhex('14AE2154A26F1D42C3C3BEDC10D05FDD9F8545BB'),
                    nickname='Test',
                ),
                'Test',
            ),
            ('$14AE2154A26F1D42C3C3BEDC10D05FDD9F8545BB~Test', 'Test'),
            ('$14AE2154A26F1D42C3C3BEDC10D05FDD9F8545BB', None),
        ],
    )
    def test_parse(self, entry, nickname):
        server = self.ADAPTER.validate_python(entry)
        assert len(server.fingerprint) == 20
        assert server.nickname == nickname

    def test_parse_error(self):
        with pytest.raises(ValueError, match='LongServerName does not start with a'):
            LongServerName.from_string('Hello world')

    @pytest.mark.parametrize(
        ('string', 'nickname'),
        [
            ('$A4DE8349C2089CC471EC12099F87BDD797EBDA8E~Test', 'Test'),
            ('$A4DE8349C2089CC471EC12099F87BDD797EBDA8E', None),
        ],
    )
    def test_serialize(self, string, nickname):
        fp = b'\xa4\xde\x83I\xc2\x08\x9c\xc4q\xec\x12\t\x9f\x87\xbd\xd7\x97\xeb\xda\x8e'
        server = LongServerName(fingerprint=fp, nickname=nickname)
        serial = LongServerNameAdapter.dump_python(server)
        assert serial == string


class TestTcpAddressPort:
    ADAPTER = TypeAdapter(TcpAddressPort)

    @pytest.mark.parametrize(
        ('entry', 'host', 'port'),
        [
            (
                TcpAddressPort(
                    host=IPv4Address('127.0.0.1'),
                    port=445,
                ),
                IPv4Address('127.0.0.1'),
                445,
            ),
            ('127.0.0.1:445', IPv4Address('127.0.0.1'), 445),
            ('[::1]:65432', IPv6Address('::1'), 65432),
        ],
    )
    def test_parse(self, entry, host, port):
        target = self.ADAPTER.validate_python(entry)
        assert target.host == host
        assert target.port == port

    @pytest.mark.parametrize(
        ('string', 'host', 'port'),
        [
            ('127.0.0.1:445', IPv4Address('127.0.0.1'), 445),
            ('[::1]:65432', IPv6Address('::1'), 65432),
        ],
    )
    def test_serialize(self, string, host, port):
        target = TcpAddressPort(host=host, port=port)
        serial = self.ADAPTER.dump_python(target)
        assert serial == string


class TestStreamTarget:
    ADAPTER = TypeAdapter(StreamTarget)

    def test_ipv4_port(self):
        line = '127.0.0.1:53'
        target = self.ADAPTER.validate_python(line)
        assert target.host == IPv4Address('127.0.0.1')
        assert target.node is None
        assert target.port == 53
        assert str(target) == line

        # Check that the target can be provided again and this is valid.
        assert self.ADAPTER.validate_python(target) == target

    def test_ipv6_port(self):
        line = '[::1]:443'
        target = self.ADAPTER.validate_python(line)
        assert target.host == IPv6Address('::1')
        assert target.node is None
        assert target.port == 443
        assert str(target) == line

    def test_domain_port(self):
        line = 'www.torproject.org:443'
        target = self.ADAPTER.validate_python(line)
        assert target.host == 'www.torproject.org'
        assert target.node is None
        assert target.port == 443
        assert str(target) == line

    def test_ipv4_exit_node(self):
        fphex = '3629DC5393D25D9588F2D613CF4185A98E405C1BFAC747F75A4B4619F47CAEA7'
        line = f'1.1.1.1.${fphex}.exit:443'
        target = self.ADAPTER.validate_python(line)
        assert target.host == IPv4Address('1.1.1.1')
        assert target.node.fingerprint.hex().upper() == fphex
        assert target.port == 443
        assert str(target) == line

    def test_ipv6_exit_node(self):
        fphex = '3629DC5393D25D9588F2D613CF4185A98E405C1BFAC747F75A4B4619F47CAEA7'
        line = f'[::1].${fphex}.exit:443'
        target = self.ADAPTER.validate_python(line)
        assert target.host == IPv6Address('::1')
        assert target.node.fingerprint.hex().upper() == fphex
        assert target.port == 443
        assert str(target) == line


class TestOnionClientAuthKey:
    ADAPTER = TypeAdapter(OnionClientAuthKey)

    def test_parse_and_encode(self):
        value = 'x25519:jPshnLNf+mpeEaBq/xEWjY5A/rnN7El8mRZmA0IyVwc'
        key = self.ADAPTER.validate_python(value)
        assert isinstance(key, X25519PrivateKey)

        serial = self.ADAPTER.dump_python(key)
        assert serial == value

    def test_user_warning(self):
        msg = 'Unhandled onion client authentication key type'
        with pytest.warns(UserWarning, match=msg):
            self.ADAPTER.dump_python('xxxx')


class TestOnionServiceKey:
    ADAPTER = TypeAdapter(OnionServiceKey)

    def test_parse_with_ed25519(self):
        value = (
            'ED25519-V3:ECum/PYnCBIHwWWmn6AaO29uY4Eq/hDEz6pLUGznA0P0ZZKoLzYbJ'
            'yURXRs0GNUz5aon9y+I3x3GauWJEXymSA=='
        )
        key = self.ADAPTER.validate_python(value)
        assert isinstance(key, OnionServiceKeyStruct)

        serial = self.ADAPTER.dump_python(key)
        assert serial == value.rstrip('=')

    def test_parse_with_rsa1024(self):
        value = (
            'RSA1024:MIICXAIBAAKBgQCdgZ2RL9T2OvYCQ6dDmiWuaxZPsL111BEDEc6HOKDw'
            'E9f9Mu4Oybd48TVMgm9/xSYLBN4gBca75fZX9oqB+umy7gNdRZyeat81YwNZUOb0'
            'Bko6Gfo6nrhssrvGETk6rRjMMRKVkeMDUQYDTF4bo2dKNxEhkmfEUbnZvjPs5E1Z'
            'RQIDAQABAoGAPVXJN02qH8zsGgugainv/JEFGjlYPjc7/LcFdxDtUzBXDumzXJze'
            'zsEXoVi19MqgOvBFU7EMKAWwPabrXxyHvqKCNoR8Iwlz7yozXchre5l87EhgReNR'
            'vGnFaUYYSrJ9vb92ytI9ZDdbf6pxd1cWxzbUghFCxpUZ4k+GmkzNka0CQQDQhLbc'
            'OOS+gc4iAa83XyXD/iJO/gYfNuObI9FOnTKVwxA6Al/TaTGp0Mk35IL1wh07EHxN'
            '7oxSGieqWriGm5ZnAkEAwV85CYRkEC7kxRlMLAjxqyVD8jh/nSvakYqqJRZ2JJdn'
            'jXO20nICowciaFKJvJSC4iU/4w3wcu9beed9ialPcwJAKYy4b0t68ScmbwpM4si3'
            '2sUSCxF9IM0sL2bEt1iFkugKnLSKabMFbWQoJFYJbnUeo/1V96V4GogRrVVkfZYV'
            'MwJAS95BcaN44wSTC2XWhfxoXR683t8d6puXIL1H7k82wTqKDWyWEVFcCXy2Gjow'
            'AkY+Z933h+0jJuUUfeq+TXGZUwJBAL5V289moEjRsqwXn8Eaw5aU/+IRBgnWxXKZ'
            '6mXD8gb5OQ7TLkcfehNHHy3JYEgfrLjZafol/9IGA7tisfp2vPs='
        )
        key = self.ADAPTER.validate_python(value)
        assert isinstance(key, RSAPrivateKey)

        serial = self.ADAPTER.dump_python(key)
        assert serial == value.rstrip('=')

    def test_serialize_ed25519(self):
        key = Ed25519PrivateKey.from_private_bytes(secrets.token_bytes(32))
        expanded = TrEd25519PrivateKey().to_expanded_bytes(key)
        expected = b64encode(expanded).decode().rstrip('=')
        result = self.ADAPTER.dump_python(key)
        assert result == f'ED25519-V3:{expected}'

    def test_user_warning(self):
        with pytest.warns(UserWarning, match='Unhandled onion service key type'):
            self.ADAPTER.dump_python('xxxx')


class TestBlockParsing:
    """Checks on our block parsing."""

    LINES: ClassVar[Sequence[str]] = [
        '-----BEGIN MESSAGE-----',
        'SbD0e0NOZ1K5Q7utFGWocGEyLGPi2KFPwniLMZb1VDymcJ7yOILYMk52kxQ+j7Cvjt/1nwdbG0Wk',
        'iBCE0R9y0oXTP9K0A2TG7sLudILSmCP35g9W3XRtYnrJNVIx3OzfKOKIo+j6SLD7xUeP1SBop3jn',
        '6s9Hd2mjPGH27gLRDbvYa3pVgUb495UnvqONbHf50SzzFe+ZsKoJSWG3jhI3Q0Db9nBwK3KLzEtc',
        'HPCshn1ZEPmMrCG3Wk4FkRb8NC6RyrY/k+Tuxem5iwb9bqkX1tqSAeP6q/7o5Q2xm6prMod/Behj',
        'vcvVixkIdFci7uCwaxfmGIUdHUa2hBviDJpX3Q==',
        '-----END MESSAGE-----',
    ]

    def test_invalid_block(self):
        with pytest.raises(ReplySyntaxError, match='Unexpected block start'):
            _parse_block(iter(self.LINES), 'PUBLIC KEY')

    def test_content_inner(self):
        inner = _parse_block(iter(self.LINES), 'MESSAGE', inner=True)
        assert inner == ''.join(self.LINES[1:-1])

    def test_content_outer(self):
        inner = _parse_block(iter(self.LINES), 'MESSAGE', inner=False)
        assert inner == '\n'.join(self.LINES)


class TestEd25519Certificate:
    """Check the parsing of ed25519 certificates."""

    #: Type adapter for Ed25519CertificateV1 certificates.
    ADAPTER: ClassVar[Ed25519CertificateV1] = TypeAdapter(Ed25519CertificateV1)

    def test_good_certificate(self):
        hexdata = (
            '010b00075d3c01d01212e107610cde2918ac3434c2142ca1105bafccd3748b21'
            'a7b58856654438010020040003e909560aa696560d7a132a17873c5098fb4186'
            '3c8535227a45d4afa3478dca2c9b5035181f8682bd7ffe5d1cc512eee000a068'
            'cf5757b3557e982194a6bb589569b76bf77bb1203ff8d87cdb38573cb770cb29'
            'db4cde0d64919554463c0302'
        )
        cert1 = self.ADAPTER.validate_python(bytes.fromhex(hexdata))
        assert isinstance(cert1, Ed25519CertificateV1)
        cert1.raise_for_invalid_signature(cert1.signing_key)

        # Also check that we can build from an already built structure.
        cert2 = self.ADAPTER.validate_python(cert1)
        assert isinstance(cert2, Ed25519CertificateV1)

        # Check certificate extensions.
        assert len(cert1.extensions) == 1

        extension = cert1.extensions[0]
        assert isinstance(extension, Ed25519CertExtensionSigningKey)
        assert extension.flags == 0

        # Simple check that an extension can be parsed again when already built.
        adapter = TypeAdapter(Ed25519CertExtension)
        extcheck = adapter.validate_python(extension)
        assert extcheck == extension

    def test_bad_certificate_version(self):
        hexdata = 'ff0b00075d3c01d01212e107610cde2918ac3434c2142ca1105bafccd3748b21'
        msg = 'Unknown ed25519 certificate version: 255'
        with pytest.raises(ReplySyntaxError, match=msg):
            self.ADAPTER.validate_python(bytes.fromhex(hexdata))

    def test_with_no_extension(self):
        hexdata = (
            '010b00075d3c01d01212e107610cde2918ac3434c2142ca1105bafccd3748b21'
            'a7b58856654438002c9b5035181f8682bd7ffe5d1cc512eee000a068cf5757b3'
            '557e982194a6bb589569b76bf77bb1203ff8d87cdb38573cb770cb29db4cde0d'
            '64919554463c0302'
        )
        cert = self.ADAPTER.validate_python(bytes.fromhex(hexdata))
        assert len(cert.extensions) == 0
        assert cert.signing_key is None

        msg = 'Ed25519 certificate has an invalid signature'
        with pytest.raises(CryptographyError, match=msg):
            cert.raise_for_invalid_signature(cert.key)

    def test_with_unknown_extension(self):
        hexdata = (
            '010b00075d3c01d01212e107610cde2918ac3434c2142ca1105bafccd3748b21'
            'a7b58856654438010002ff01abcd2c9b5035181f8682bd7ffe5d1cc512eee000'
            'a068cf5757b3557e982194a6bb589569b76bf77bb1203ff8d87cdb38573cb770'
            'cb29db4cde0d64919554463c0302'
        )
        cert = self.ADAPTER.validate_python(bytes.fromhex(hexdata))
        assert len(cert.extensions) == 1
        assert cert.signing_key is None

        # Check that extension.
        extension = cert.extensions[0]
        assert isinstance(extension, Ed25519CertExtensionUnkown)
        assert extension.data.hex() == 'abcd'

        msg = 'Ed25519 certificate has an unknown extension affecting validation'
        with pytest.raises(CryptographyError, match=msg):
            cert.raise_for_invalid_signature(None)


class TestTrLinkSpecifierList:
    """Test a list of link specifiers."""

    ADAPTER = TypeAdapter(Annotated[Sequence[LinkSpecifierStruct], TrLinkSpecifierList()])

    def test_parse(self):
        hexdata = (
            '040006416c88b70050021415291291d81e404fb6bec16d366608590d2b024703'
            '20db2ff9542d8472371105cbce8e8fba91196d69d25aa7bacd7c7677fdb40771'
            '0001122a0104f9006b340800000000000000040050'
        )
        links = self.ADAPTER.validate_python(bytes.fromhex(hexdata))
        assert len(links) == 4
        assert links[0].type == 0
        assert self.ADAPTER.validate_python(links) == links


class TestHsDescV3Layer2:
    """Test layer2 parsing."""

    ADAPTER = TypeAdapter(HsDescV3Layer2)

    def test_no_introduction_point(self):
        lines = [
            'create2-formats 2',
            'flow-control 1-2 31',
        ]
        mapping = HsDescV3Layer2.text_to_mapping('\n'.join(lines))
        layer2 = self.ADAPTER.validate_python(mapping)
        assert len(layer2.introduction_points) == 0

    def test_with_flow_control(self):
        lines = [
            'create2-formats 2',
            'flow-control 1-2 31',
        ]
        mapping = HsDescV3Layer2.text_to_mapping('\n'.join(lines))
        layer2 = self.ADAPTER.validate_python(mapping)
        assert layer2.flow_control.sendme_inc == 31
        assert layer2.flow_control.version_range.min == 1
        assert layer2.flow_control.version_range.max == 2
        assert layer2.single_service is False

    def test_with_good_intro_auth_required(self):
        lines = [
            'create2-formats 2',
            'intro-auth-required ed25519',
            'single-onion-service',
        ]
        mapping = HsDescV3Layer2.text_to_mapping('\n'.join(lines))
        layer2 = self.ADAPTER.validate_python(mapping)
        assert layer2.single_service is True
        assert layer2.introduction_auth == {'ed25519'}
        assert layer2.pow_params is None

    def test_with_pow_params(self):
        lines = [
            'create2-formats 2',
            'pow-params v1 FD+Ai7r8Xx1BhF42JbtJkg 2 1738188453',
        ]
        mapping = HsDescV3Layer2.text_to_mapping('\n'.join(lines))
        layer2 = self.ADAPTER.validate_python(mapping)
        assert layer2.pow_params.type == 'v1'
        assert layer2.pow_params.suggested_effort == 2
        assert isinstance(layer2.pow_params.expiration, datetime)



from __future__ import annotations

import asyncio
import logging
from asyncio import Condition, Task
from contextlib import AsyncExitStack, suppress
from dataclasses import dataclass
from typing import TYPE_CHECKING, ClassVar

from pydantic import NonNegativeInt

from .event import Event, EventNetworkLiveness, EventStatusClient, EventWord
from .exceptions import ControllerError, ReplyStatusError
from .structures import Signal, StatusActionClient, StatusClientBootstrap
from .utils import Message

if TYPE_CHECKING:
    from types import TracebackType
    from typing import Self

    from .controller import Controller

logger = logging.getLogger(__package__)


@dataclass(slots=True)
class ControllerStatus:
    """
    Keep track of the Controller's status.

    It is used to keep track of various status parameters from Tor's daemon.
    """

    #: Tor's bootstrap progress status (in percent).
    bootstrap: NonNegativeInt = 0

    #: Whether Tor has established circuits.
    has_circuits: bool = False

    #: Whether Tor has enough directory information.
    has_dir_info: bool = False

    #: Whether Tor has a working network.
    net_liveness: bool = False

    @property
    def is_healthy(self) -> bool:
        """
        Tell whether we are healthy enough to run workers.

        Returns:
            A boolean value that tells whether we are healthy or not.

        """
        return bool(
            self.bootstrap == 100
            and self.has_dir_info
            and (self.has_circuits or self.net_liveness),
        )


class Monitor:
    """Monitor controller's status."""

    DEFAULT_DORMANT_TIMEOUT: ClassVar[float] = 24 * 3600.0

    def __init__(
        self,
        controller: Controller,
        *,
        keepalive: bool = True,
    ) -> None:
        """
        Create a new instance of a monitor to check for Tor's daemon status.

        Args:
            controller: a controller connected to Tor's daemon

        Keyword Args:
            keepalive: whether we should run a task to keep Tor in ``ACTIVE`` mode.

        """
        self._context = None  # type: AsyncExitStack | None
        self._condition = Condition()
        self._controller = controller
        self._do_keepalive = keepalive
        self._status = ControllerStatus()

    async def __aenter__(self) -> Self:
        """
        Start the controller's monitoring.

        This subscribes to relevant events from the controller and optionally starts
        the keep-alive task used to ensure that Tor always stay ``ACTIVE``. This later
        part is important if you don't plan on using the socks port but will only
        use the control port.

        Raises:
            RuntimeError: when the controller has already been entered

        Returns:
            The exact same :class:`Monitor` object.

        """
        if self.is_entered:
            msg = 'Monitor is already running!'
            raise RuntimeError(msg)

        controller = self._controller
        context = await AsyncExitStack().__aenter__()
        try:
            # Get notified when a 'STATUS_CLIENT' event occurs.
            await controller.add_event_handler(
                EventWord.STATUS_CLIENT,
                self._on_ctrl_client_status,
            )
            context.push_async_callback(
                controller.del_event_handler,
                EventWord.STATUS_CLIENT,
                self._on_ctrl_client_status,
            )

            # Get notified when a 'NETWORK_LIVENESS' event occurs.
            await controller.add_event_handler(
                EventWord.NETWORK_LIVENESS,
                self._on_ctrl_liveness_status,
            )
            context.push_async_callback(
                controller.del_event_handler,
                EventWord.NETWORK_LIVENESS,
                self._on_ctrl_liveness_status,
            )

            if self._do_keepalive:  # pragma: no branch
                # Only enable the keep alive task when it is needed.
                # This option was introduced in Tor `0.4.6.2`.
                try:
                    reply = await self._controller.get_conf('DormantTimeoutEnabled')
                    dormant = bool(reply.get('DormantTimeoutEnabled', True))
                except ReplyStatusError:  # pragma: no cover
                    dormant = True

                if dormant:  # pragma: no branch
                    task_keepalive = asyncio.create_task(
                        self._keepalive_task(),
                        name='aiostem.monitor.keepalive',
                    )

                    async def cancel_keepalive(task: Task[None]) -> None:
                        task.cancel('Monitor is closing')
                        await asyncio.gather(task, return_exceptions=True)

                    context.push_async_callback(cancel_keepalive, task_keepalive)

            await self._fetch_controller_status()
        except BaseException:  # pragma: no cover
            await context.aclose()
            raise
        else:
            self._context = context
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool:
        """
        Exit the monitor's context and unregister from the underlying events.

        Returns:
            :obj:`False` to let any exception flow through the call stack.

        """
        context = self._context
        try:
            if context is not None:
                with suppress(ControllerError):
                    await context.__aexit__(exc_type, exc_val, exc_tb)
        finally:
            self._context = None

        # Do not prevent the original exception from going further.
        return False

    async def _keepalive_task(self) -> None:
        """
        Keep the Tor daemon in an `ALIVE` state.

        This is needed when the user does not plan on using the socks socket
        but still needs to send commands regularly through the control port.
        """
        dormant_timeout = self.DEFAULT_DORMANT_TIMEOUT

        reply = await self._controller.get_conf('DormantClientTimeout')
        if reply.is_success:  # pragma: no branch
            value = reply.get('DormantClientTimeout')
            if isinstance(value, str):  # pragma: no branch
                dormant_timeout = float(value)
            logger.debug("Config 'DormantClientTimeout' is set to %d", value)

        delay = 0.95 * dormant_timeout
        while True:
            logger.info("Sending the 'ACTIVE' signal to the controller")
            await self._controller.signal(Signal.ACTIVE)
            await asyncio.sleep(delay)

    async def _fetch_controller_status(self) -> bool:
        """
        Fetch and update our view of the controller's status.

        Returns:
            Whether the underlying Tor daemon is healthy.

        """
        reply = await self._controller.get_info(
            'network-liveness',
            'status/bootstrap-phase',
            'status/circuit-established',
            'status/enough-dir-info',
        )
        reply.raise_for_status()

        values = {}  # type: dict[str, str]
        for key, val in reply.items():
            if isinstance(key, str) and isinstance(val, str):  # pragma: no branch
                values[key] = val

        # Build a fake message to create an event-like object.
        # This avoids all the manual parsing of this thing...
        message = Message(
            header='STATUS_CLIENT ' + values['status/bootstrap-phase'],
            status=650,
        )

        status = EventStatusClient.from_message(message)
        if isinstance(status.arguments, StatusClientBootstrap):  # pragma: no cover
            self._status.bootstrap = status.arguments.progress

        self._status.has_circuits = bool(values['status/circuit-established'] == '1')
        self._status.has_dir_info = bool(values['status/enough-dir-info'] == '1')
        self._status.net_liveness = bool(values['network-liveness'] == 'up')

        return self.is_healthy

    async def _on_ctrl_liveness_status(self, event: Event) -> None:
        """Handle a `NETWORK_LIVENESS` event."""
        if isinstance(event, EventNetworkLiveness):  # pragma: no branch
            async with self._condition:
                logger.debug('Network liveness: %s', event.status)
                self._status.net_liveness = bool(event.status)
                self._condition.notify_all()

    async def _on_ctrl_client_status(self, event: Event) -> None:
        """
        Handle a `STATUS_CLIENT` event.

        Note that this is an event handler executed in the receive loop from the controller.
        You cannot perform new controller requests from here, use a queue or something else.
        """
        if isinstance(event, EventStatusClient):  # pragma: no branch
            async with self._condition:
                match event.action:
                    case StatusActionClient.BOOTSTRAP:
                        args = event.arguments
                        if isinstance(args, StatusClientBootstrap):  # pragma: no cover
                            self._status.bootstrap = args.progress
                    case StatusActionClient.CIRCUIT_ESTABLISHED:
                        self._status.has_circuits = True
                    case StatusActionClient.CIRCUIT_NOT_ESTABLISHED:
                        self._status.has_circuits = False
                    case StatusActionClient.ENOUGH_DIR_INFO:
                        self._status.has_dir_info = True
                    case StatusActionClient.NOT_ENOUGH_DIR_INFO:
                        self._status.has_dir_info = False
                    case _:
                        # We are not not interested in other events.
                        # This also means that our status hasn't changed.
                        return

                # Maybe we need to do something about the current working state.
                logger.debug('ClientStatus: %s %s', event.action, event.arguments)
                self._condition.notify_all()

    @property
    def condition(self) -> Condition:
        """
        Get the underlying asyncio condition.

        This condition is triggered when any update has been received, which does
        not guarantee that the global health status has changed.

        USE EXAMPLE::

           async with monitor.condition:
               await monitor.condition.wait_for(lambda: monitor.is_healthy)
               print('Monitor is now healthy.')

        """
        return self._condition

    @property
    def is_entered(self) -> bool:
        """Tell whether the monitor context is currently entered."""
        return bool(self._context is not None)

    @property
    def is_healthy(self) -> bool:
        """
        Tell whether the underlying controller is healthy.

        This is a shorthand for ``monitor.status.is_healthy``.

        """
        return self._status.is_healthy

    @property
    def status(self) -> ControllerStatus:
        """
        Get the current controller status.

        Note that this status is no a copy but a reference to the internal status.
        Any change on its field values have consequences on the Monitor's view.

        """
        return self._status

    async def wait_for_error(self) -> ControllerStatus:
        """
        Wait until the controller stops being healthy.

        This method can be implemented using :attr:`condition`, :attr:`is_healthy` and
        :attr:`status`.

        Returns:
            The current controller's status.

        """
        async with self._condition:
            await self._condition.wait_for(lambda: not self._status.is_healthy)
        return self._status

    async def wait_until_healthy(self) -> ControllerStatus:
        """
        Wait until the controller is ready and healthy.

        This method can be implemented using :attr:`condition`, :attr:`is_healthy` and
        :attr:`status`.

        Returns:
            The current controller's status.

        """
        async with self._condition:
            await self._condition.wait_for(lambda: self._status.is_healthy)
        return self._status



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Sequence
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Final

from ..exceptions import ProtocolError

if TYPE_CHECKING:
    from asyncio import StreamReader
    from collections.abc import AsyncIterator


@dataclass(frozen=True, kw_only=True)
class BaseMessage(ABC):
    """
    Base class for any kind of message items.

    A received message (like a reply) can contain sub-messages that also inherit
    from this base class.

    """

    #: The end of line applied while serializing messages.
    END_OF_LINE: Final[str] = '\r\n'

    #: Status code of this message or message item.
    status: int
    #: Text that comes along the status in this (sub-)message.
    header: str

    @property
    def is_error(self) -> bool:
        """Whether our status is an error (greater or equal to 400)."""
        return bool(self.status >= 400 and self.status != 650)

    @property
    def is_event(self) -> bool:
        """
        Tell whether this message is an event.

        This property is a simple helper to tell whether our status equals to 650.

        """
        return bool(self.status == 650)

    @property
    def is_success(self) -> bool:
        """Whether our status is a success status (=250)."""
        return bool(self.status == 250)

    @property
    def keyword(self) -> str:
        """
        Extract the first word from the header.

        Note:
            This is not relevant for all kind of messages.

        """
        return self.header.split(' ', maxsplit=1)[0]

    @abstractmethod
    def serialize(self) -> str:
        """Serialize this message to text that could have been sent."""


@dataclass(frozen=True, kw_only=True)
class MessageData(BaseMessage):
    """A sub-message with a body part attached."""

    #: Additional data (body of the message).
    data: str = ''

    def serialize(self) -> str:
        """Serialize this data sub-message to a string."""
        lines = [f'{self.status:03d}+{self.header}']
        for line in self.data.splitlines():
            if line.startswith('.'):
                line = '.' + line
            lines.append(line)
        lines.append('.')
        return self.END_OF_LINE.join(lines) + self.END_OF_LINE


@dataclass(frozen=True, kw_only=True)
class MessageLine(BaseMessage):
    """A sub-message with only a single line."""

    def serialize(self) -> str:
        """Serialize this line sub-message to a string."""
        return f'{self.status:03d}-{self.header}{self.END_OF_LINE}'


@dataclass(frozen=True, kw_only=True)
class Message(BaseMessage):
    """Utility class used to parse any received message."""

    #: List of sub-messages received within this message.
    items: Sequence[MessageLine | MessageData] = field(default_factory=list)

    @property
    def keyword(self) -> str:
        """
        Extract the first word from the header.

        Note:
            An event always provide its keyword in the fist item.

        """
        if self.is_event and len(self.items) > 0:
            return self.items[0].keyword
        return super().keyword

    def serialize(self) -> str:
        """Serialize this message to a string."""
        text = ''
        for item in self.items:
            text += item.serialize()
        text += f'{self.status:03d} {self.header}{self.END_OF_LINE}'
        return text


async def messages_from_stream(stream: StreamReader) -> AsyncIterator[Message]:
    """
    Parse messages from the underlying stream.

    Args:
        stream: The asyncio stream reader to read messages from.

    Raises:
        ProtocolError: When we receive a malformed message.

    Yields:
        Messages as they are parsed.

    """
    items = []  # type: list[MessageLine | MessageData]
    lines = []  # type: list[str]
    data = None  # type: MessageData | None

    while line_bytes := await stream.readline():
        line = line_bytes.decode('ascii').removesuffix(BaseMessage.END_OF_LINE)

        # Continuation of a data sub-message parser.
        if isinstance(data, MessageData):
            if line == '.':
                items.append(
                    MessageData(
                        status=data.status,
                        header=data.header,
                        data='\n'.join(lines),
                    )
                )
                lines.clear()
                data = None
            else:
                if line.startswith('.'):
                    line = line[1:]
                lines.append(line)
            continue

        # We expect a valid header, and require at least 4 characters.
        if len(line) < 4:
            msg = f"Received line is too short: '{line}'"
            raise ProtocolError(msg)

        # The first three characters form a decimal status code.
        try:
            status = int(line[0:3], 10)
        except ValueError:
            msg = f"Unable to parse status code on line '{line}'"
            raise ProtocolError(msg) from None

        separator = line[3]
        content = line[4:]

        match separator:
            case ' ':
                yield Message(status=status, header=content, items=[*items])
                items.clear()

            case '-':
                items.append(MessageLine(status=status, header=content))

            case '+':
                data = MessageData(status=status, header=content)

            case _:
                msg = f"Unrecognized separator on line '{line}'"
                raise ProtocolError(msg)
